# Copyright 2017, 2018, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from abc import ABCMeta, abstractmethod
from digi.xbee.packets.aft import ApiFrameType
from digi.xbee.models.atcomm import SpecialByte
from digi.xbee.models.mode import OperatingMode
from digi.xbee.exception import InvalidPacketException, InvalidOperatingModeException
from digi.xbee.util import utils
from enum import Enum, unique


@unique
class DictKeys(Enum):
    """
    This enumeration contains all keys used in dictionaries returned by ``to_dict()``
    method of :class:`.XBeePacket`.
    """

    HEADER_BYTE = "header"
    LENGTH = "length"
    FRAME_TYPE = "fr_type"
    FRAME_ID = "fr_id"
    FRAME_SPEC_DATA = "fr_spec_data"
    API_DATA = "api_spec_data"
    RF_DATA = "rf_data"
    CHECKSUM = "checksum"
    COMMAND = "at_command"
    PARAMETER = "parameter"
    X64BIT_ADDR = "x64_addr"
    X16BIT_ADDR = "x16_addr"
    TRANSMIT_OPTIONS = "transmit_options"
    RECEIVE_OPTIONS = "receive_options"
    BROADCAST_RADIUS = "broadcast_radius"
    TRANS_R_COUNT = "transmit_retry_count"
    TS_STATUS = "ts_status"
    DS_STATUS = "ds_status"
    AT_CMD_STATUS = "at_command_status"
    MODEM_STATUS = "modem_status"
    NUM_SAMPLES = "num_samples"
    DIGITAL_MASK = "digital_mask"
    ANALOG_MASK = "analog_mask"
    RSSI = "rssi"
    SOURCE_ENDPOINT = "source_endpoint"
    DEST_ENDPOINT = "dest_endpoint"
    CLUSTER_ID = "cluster_id"
    PROFILE_ID = "profile_id"
    PHONE_NUMBER = "phone_number"
    DEST_IPV4_ADDR = "dest_ipv4_address"
    SRC_IPV4_ADDR = "source_ipv4_address"
    DEST_PORT = "dest_port"
    SRC_PORT = "source_port"
    IP_PROTOCOL = "ip_protocol"
    STATUS = "status"
    TRANSPORT = "transport"
    FLAGS = "flags"
    REQUEST_ID = "request_id"
    TARGET = "target"
    RESERVED = "reserved"
    DC_STATUS = "device_cloud_status"
    FRAME_ERROR = "frame_error"
    PATH_LENGTH = "path_length"
    PATH = "path"
    CONTENT_TYPE_LENGTH = "content_type_length"
    CONTENT_TYPE = "content_type"


class XBeePacket:
    """
    This abstract class represents the basic structure of an XBee packet.
 
    Derived classes should implement their own payload generation depending on their type.
    
    Generic actions like checksum compute or packet length calculation is performed here.
    """

    __HASH_SEED = 23

    __metaclass__ = ABCMeta
    __ESCAPE_BYTES = [i.value for i in SpecialByte]
    __ESCAPE_FACTOR = 0x20
    ESCAPE_BYTE = SpecialByte.ESCAPE_BYTE.code

    def __init__(self):
        """
        Class constructor. Instantiates a new :class:`.XBeePacket` object.
        """
        pass

    def __len__(self):
        """
        Returns the length value of the XBeePacket.
        
        The length is the number of bytes between the length field and the checksum field.
        
        Returns:
            Integer: length value of the XBeePacket.
            
        .. seealso::
           | :mod:`.factory`
        """
        return len(self.get_frame_spec_data())

    def __str__(self):
        """
        Returns the packet information as dictionary.
        
        Returns:
            Dictionary: the packet information.
        """
        return str(self.to_dict())

    def __eq__(self, other):
        """
        Returns whether the given object is equal to this one.

        Args:
            other: the object to compare.

        Returns:
            Boolean: ``True`` if the objects are equal, ``False`` otherwise.
        """
        if not isinstance(other, XBeePacket):
            return False
        return other.output() == self.output()

    def __hash__(self):
        """
        Returns a hash code value for the object.

        Returns:
            Integer: hash code value for the object.
        """
        res = self.__HASH_SEED
        for b in self.output():
            res = 31 * (res + b)
        return res

    def get_checksum(self):
        """
        Returns the checksum value of this XBeePacket.
        
        The checksum is the last 8 bits of the sum of the bytes between the length field and the checksum field.
        
        Returns:
            Integer: checksum value of this XBeePacket.
        
        .. seealso::
           | :mod:`.factory`
        """
        return 0xFF - (sum(self.get_frame_spec_data()) & 0xFF)

    def output(self, escaped=False):
        """
        Returns the raw bytearray of this XBeePacket, ready to be send by the serial port.
        
        Args:
            escaped (Boolean): indicates if the raw bytearray will be escaped or not.
        
        Returns:
            Bytearray: raw bytearray of the XBeePacket.
        """
        frame = self.__build_complete_frame_without_header(self.get_frame_spec_data())
        if escaped:
            frame = self._escape_data(frame)
        frame.insert(0, SpecialByte.HEADER_BYTE.code)
        return frame

    def to_dict(self):
        """
        Returns a dictionary with all information of the XBeePacket fields.
        
        Returns:
            Dictionary: dictionary with all information of the XBeePacket fields.
        """
        return {DictKeys.HEADER_BYTE:     SpecialByte.HEADER_BYTE.code,
                DictKeys.LENGTH:          len(self),
                DictKeys.FRAME_SPEC_DATA: self._get_frame_spec_data_dict(),
                DictKeys.CHECKSUM:        self.get_checksum()}

    @staticmethod
    def create_packet(raw, operating_mode):
        """
        Abstract method.
        Creates a full XBeePacket with the given parameters.
        This function ensures that the XBeePacket returned is valid and is well built (if not exceptions are raised).
        
        If _OPERATING_MODE is API2 (API escaped) this method des-escape 'raw' and build the XBeePacket.
        Then, you can use :meth:`.XBeePacket.output` to get the escaped bytearray or not escaped.
        
        Args:
            raw (Bytearray): bytearray with which the frame will be built. Must be a full frame
                represented by a bytearray.
            operating_mode (:class:`.OperatingMode`): The mode in which the frame ('byteArray') was captured.
            
        Returns:
            :class:`.XBeePacket`: the XBee packet created.
        
        Raises:
            InvalidPacketException: if something is wrong with ``raw`` and the packet cannot be built well.
        """
        raise NotImplementedError("Implement this function.")

    @abstractmethod
    def get_frame_spec_data(self):
        """
        Returns the data between the length field and the checksum field as bytearray.
        This data is never escaped.
        
        Returns:
            Bytearray: the data between the length field and the checksum field as bytearray.
        
        .. seealso::
           | :mod:`.factory`
        """
        pass

    @abstractmethod
    def _get_frame_spec_data_dict(self):
        """
        Similar to :meth:`.XBeePacket.get_frame_spec_data` but returns the data as dictionary.
        
        Returns:
            Dictionary: the data between the length field and the checksum field as bytearray.
        """
        pass

    @staticmethod
    def _escape_data(data):
        """
        Escapes the bytearray 'data'.

        Args:
            data (Bytearray): the bytearray to escape.

        Returns:
            Bytearray: 'data' escaped.
        """
        esc_data = bytearray()
        for i in data:
            if i in XBeePacket.__ESCAPE_BYTES:
                esc_data.append(SpecialByte.ESCAPE_BYTE.code)
                esc_data.append(i ^ XBeePacket.__ESCAPE_FACTOR)
            else:
                esc_data.append(i)
        return esc_data

    @staticmethod
    def unescape_data(data):
        """
        Un-escapes the provided bytearray data.

        Args:
            data (Bytearray): the bytearray to unescape.

        Returns:
            Bytearray: ``data`` unescaped.
        """
        new_data = bytearray(0)
        des_escape = False
        for byte in data:
            if byte == XBeePacket.ESCAPE_BYTE:
                des_escape = True
            else:
                new_data.append(byte ^ 0x20 if des_escape else byte)
                des_escape = False
        return new_data

    def __build_complete_frame_without_header(self, frame_spec_data):
        """
        Builds a complete non-escaped frame from the given frame specific data.
        Complete frame is: Start delimiter + length + frame specific data + checksum.
        
        Args:
            frame_spec_data (Bytearray): the frame specific data.
            
        Returns:
            Bytearray: the complete frame as bytearray.
        """
        frame = utils.int_to_length(len(frame_spec_data)) + frame_spec_data
        frame.append(self.get_checksum())
        return frame


class XBeeAPIPacket(XBeePacket):
    """
    This abstract class provides the basic structure of a API frame.
    
    Derived classes should implement their own methods to generate the API 
    data and frame ID in case they support it.
    
    Basic operations such as frame type retrieval are performed in this class.
    
    .. seealso::
       | :class:`.XBeePacket`
    """
    __metaclass__ = ABCMeta

    def __init__(self, api_frame_type):
        """
        Class constructor. Instantiates a new :class:`.XBeeAPIPacket` object with the provided parameters.
        
        Args:
            api_frame_type (:class:`.ApiFrameType` or Integer): The API frame type.

        .. seealso::
           | :class:`.ApiFrameType`
           | :class:`.XBeePacket`
        """
        super(XBeeAPIPacket, self).__init__()
        # Check the type of the API frame type.
        if isinstance(api_frame_type, ApiFrameType):
            self._frame_type = api_frame_type
            self._frame_type_value = api_frame_type.code
        else:
            self._frame_type = ApiFrameType.get(api_frame_type)
            self._frame_type_value = api_frame_type
        self._frame_id = 0

    def get_frame_spec_data(self):
        """
        Override method.
        
        .. seealso::
           | :meth:`.XBeePacket.get_frame_spec_data`
        """
        data = self._get_api_packet_spec_data()
        if self.needs_id():
            data.insert(0, self._frame_id)
        data.insert(0, self._frame_type_value)
        return data

    def get_frame_type(self):
        """
        Returns the frame type of this packet.
        
        Returns:
            :class:`.ApiFrameType`: the frame type of this packet.

        .. seealso::
           | :class:`.ApiFrameType`
        """
        return self._frame_type

    def get_frame_type_value(self):
        """
        Returns the frame type integer value of this packet.

        Returns:
            Integer: the frame type integer value of this packet.

        .. seealso::
           | :class:`.ApiFrameType`
        """
        return self._frame_type_value

    def _get_frame_spec_data_dict(self):
        """
        Override method.
        
        .. seealso::
           | :meth:`.XBeePacket.get_frame_spec_data_dict`
        """
        return {DictKeys.FRAME_TYPE: self.get_frame_type(),
                DictKeys.FRAME_ID:   self._frame_id if self.needs_id() else "NO ID",
                DictKeys.API_DATA:   self._get_api_packet_spec_data_dict()}

    def is_broadcast(self):
        """
        Returns whether this packet is broadcast or not.
        
        Returns:
            Boolean: ``True`` if this packet is broadcast, ``False`` otherwise.
        """
        return False

    def __get_frame_id(self):
        """
        Returns the frame ID of the packet.

        Returns:
            Integer: the frame ID of the packet.
        """
        return self._frame_id

    def __set_frame_id(self, frame_id):
        """
        Sets the frame ID of the packet.

        Args:
            frame_id (Integer): the new frame ID of the packet. Must be between 0 and 255.

        Raises:
            ValueError: if ``frame_id`` is less than 0 or greater than 255.
        """
        if frame_id < 0 or frame_id > 255:
            raise ValueError("Frame ID must be between 0 and 255.")
        self._frame_id = frame_id

    @staticmethod
    def _check_api_packet(raw, min_length=5):
        """
        Checks the not escaped  bytearray 'raw' meets conditions.
        
        Args:
            raw (Bytearray): non-escaped bytearray to be checked.
            min_length (Integer): the minimum length of the packet in bytes.
            
        Raises:
            InvalidPacketException: if the bytearray length is less than 5. (start delim. + length (2 bytes) +
                frame type + checksum = 5 bytes).
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field:
                bytes 2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
        
        .. seealso::
           | :mod:`.factory`
        """
        if len(raw) < min_length:
                raise InvalidPacketException("Bytearray must have, at least, 5 of complete length (header, length, "
                                             "frameType, checksum)")

        if raw[0] & 0xFF != SpecialByte.HEADER_BYTE.code:
            raise InvalidPacketException("Bytearray must start with the header byte (SpecialByte.HEADER_BYTE.code)")

        # real frame specific data length
        real_length = len(raw[3:-1])
        # length is specified in the length field.
        length_field = utils.length_to_int(raw[1:3])
        if real_length != length_field:
            raise InvalidPacketException("The real length of this frame is distinct than the specified by length "
                                         "field (bytes 2 and 3)")

        if 0xFF - (sum(raw[3:-1]) & 0xFF) != raw[-1]:
            raise InvalidPacketException("Wrong checksum")

    @abstractmethod
    def _get_api_packet_spec_data(self):
        """
        Returns the frame specific data without frame type and frame ID fields.
        
        Returns:
            Bytearray: the frame specific data without frame type and frame ID fields.
        """
        pass

    @abstractmethod
    def needs_id(self):
        """
        Returns whether the packet requires frame ID or not.
        
        Returns:
            Boolean: ``True`` if the packet needs frame ID, ``False`` otherwise.
        """
        pass

    @abstractmethod
    def _get_api_packet_spec_data_dict(self):
        """
        Similar to :meth:`XBeeAPIPacket._get_api_packet_spec_data` but returns data as dictionary or list.
        
        Returns:
            Dictionary: data as dictionary or list.
        """
        pass

    frame_id = property(__get_frame_id, __set_frame_id)


class GenericXBeePacket(XBeeAPIPacket):
    """
    This class represents a basic and Generic XBee packet.

    .. seealso::
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 5

    def __init__(self, rf_data):
        """
        Class constructor. Instantiates a :class:`.GenericXBeePacket` object with the provided parameters.
        
        Args:
            rf_data (bytearray): the frame specific data without frame type and frame ID.
        
        .. seealso::
           | :mod:`.factory`
           | :class:`.XBeeAPIPacket`
        """
        super(GenericXBeePacket, self).__init__(api_frame_type=ApiFrameType.GENERIC)
        self.__rf_data = rf_data

    @staticmethod
    def create_packet(raw, operating_mode=OperatingMode.API_MODE):
        """
        Override method.

        Returns:
            :class:`.GenericXBeePacket`: the GenericXBeePacket generated.
            
        Raises:
            InvalidPacketException: if the bytearray length is less than 5. (start delim. + length (2 bytes) +
                frame type + checksum = 5 bytes).
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field: bytes
                2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidPacketException: if the frame type is different than :attr:`.ApiFrameType.GENERIC`.
            InvalidOperatingModeException: if ``operating_mode`` is not supported.

        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=GenericXBeePacket.__MIN_PACKET_LENGTH)

        if raw[3] != ApiFrameType.GENERIC.code:
            raise InvalidPacketException("Wrong frame type, expected: " + ApiFrameType.GENERIC.description +
                                         ". Value: " + ApiFrameType.GENERIC.code)

        return GenericXBeePacket(raw[4:-1])

    def _get_api_packet_spec_data(self):
        """
        Override method.
        
        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data`
        """
        return bytearray(self.__rf_data)

    def needs_id(self):
        """
        Override method.
        
        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return False

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.
        
        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data_dict`
        """
        return {DictKeys.RF_DATA: self.__rf_data}


class UnknownXBeePacket(XBeeAPIPacket):
    """
    This class represents an unknown XBee packet.

    .. seealso::
       | :class:`.XBeeAPIPacket`
    """

    __MIN_PACKET_LENGTH = 5

    def __init__(self, api_frame, rf_data):
        """
        Class constructor. Instantiates a :class:`.UnknownXBeePacket` object with the provided parameters.

        Args:
            api_frame (Integer): the API frame integer value of this packet.
            rf_data (bytearray): the frame specific data without frame type and frame ID.

        .. seealso::
           | :mod:`.factory`
           | :class:`.XBeeAPIPacket`
        """
        super(UnknownXBeePacket, self).__init__(api_frame_type=api_frame)
        self.__rf_data = rf_data

    @staticmethod
    def create_packet(raw, operating_mode=OperatingMode.API_MODE):
        """
        Override method.

        Returns:
            :class:`.UnknownXBeePacket`: the UnknownXBeePacket generated.

        Raises:
            InvalidPacketException: if the bytearray length is less than 5. (start delim. + length (2 bytes) +
                frame type + checksum = 5 bytes).
            InvalidPacketException: if the length field of 'raw' is different than its real length. (length field: bytes
                2 and 3)
            InvalidPacketException: if the first byte of 'raw' is not the header byte. See :class:`.SpecialByte`.
            InvalidPacketException: if the calculated checksum is different than the checksum field value (last byte).
            InvalidOperatingModeException: if ``operating_mode`` is not supported.

        .. seealso::
           | :meth:`.XBeePacket.create_packet`
           | :meth:`.XBeeAPIPacket._check_api_packet`
        """
        if operating_mode != OperatingMode.ESCAPED_API_MODE and operating_mode != OperatingMode.API_MODE:
            raise InvalidOperatingModeException(operating_mode + " is not supported.")

        XBeeAPIPacket._check_api_packet(raw, min_length=UnknownXBeePacket.__MIN_PACKET_LENGTH)

        return UnknownXBeePacket(raw[3], raw[4:-1])

    def _get_api_packet_spec_data(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data`
        """
        return bytearray(self.__rf_data)

    def needs_id(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket.needs_id`
        """
        return False

    def _get_api_packet_spec_data_dict(self):
        """
        Override method.

        .. seealso::
           | :meth:`.XBeeAPIPacket._get_api_packet_spec_data_dict`
        """
        return {DictKeys.RF_DATA: self.__rf_data}



# Copyright 2017, Digi International Inc.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import time

from digi.xbee.models.status import NetworkDiscoveryStatus
from digi.xbee.devices import XBeeDevice

# TODO: Replace with the serial port where your local module is connected to.
PORT = "COM1"
# TODO: Replace with the baud rate of your local module.
BAUD_RATE = 9600


def main():
    print(" +---------------------------------------------+")
    print(" | XBee Python Library Discover Devices Sample |")
    print(" +---------------------------------------------+\n")

    device = XBeeDevice(PORT, BAUD_RATE)

    try:
        device.open()

        xbee_network = device.get_network()

        xbee_network.set_discovery_timeout(15)  # 15 seconds.

        xbee_network.clear()

        # Callback for discovered devices.
        def callback_device_discovered(remote):
            print("Device discovered: %s" % remote)

        # Callback for discovery finished.
        def callback_discovery_finished(status):
            if status == NetworkDiscoveryStatus.SUCCESS:
                print("Discovery process finished successfully.")
            else:
                print("There was an error discovering devices: %s" % status.description)

        xbee_network.add_device_discovered_callback(callback_device_discovered)

        xbee_network.add_discovery_process_finished_callback(callback_discovery_finished)

        xbee_network.start_discovery_process()

        print("Discovering remote XBee devices...")

        while xbee_network.is_discovery_running():
            time.sleep(0.1)

    finally:
        if device is not None and device.is_open():
            device.close()


if __name__ == '__main__':
    main()



# Copyright 2017, 2018, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from enum import Enum, unique
from digi.xbee.util import utils


@unique
class ApiFrameType(Enum):
    """
    This enumeration lists all the available frame types used in any XBee protocol.
    
    | Inherited properties:
    |     **name** (String): the name (id) of this ApiFrameType.
    |     **value** (String): the value of this ApiFrameType.
    
    """
    TX_64 = (0x00, "TX (Transmit) Request 64-bit address")
    TX_16 = (0x01, "TX (Transmit) Request 16-bit address")
    REMOTE_AT_COMMAND_REQUEST_WIFI = (0x07, "Remote AT Command Request (Wi-Fi)")
    AT_COMMAND = (0x08, "AT Command")
    AT_COMMAND_QUEUE = (0x09, "AT Command Queue")
    TRANSMIT_REQUEST = (0x10, "Transmit Request")
    EXPLICIT_ADDRESSING = (0x11, "Explicit Addressing Command Frame")
    REMOTE_AT_COMMAND_REQUEST = (0x17, "Remote AT Command Request")
    TX_SMS = (0x1F, "TX SMS")
    TX_IPV4 = (0x20, "TX IPv4")
    SEND_DATA_REQUEST = (0x28, "Send Data Request")
    DEVICE_RESPONSE = (0x2A, "Device Response")
    RX_64 = (0x80, "RX (Receive) Packet 64-bit Address")
    RX_16 = (0x81, "RX (Receive) Packet 16-bit Address")
    RX_IO_64 = (0x82, "IO Data Sample RX 64-bit Address Indicator")
    RX_IO_16 = (0x83, "IO Data Sample RX 16-bit Address Indicator")
    REMOTE_AT_COMMAND_RESPONSE_WIFI = (0x87, "Remote AT Command Response (Wi-Fi)")
    AT_COMMAND_RESPONSE = (0x88, "AT Command Response")
    TX_STATUS = (0x89, "TX (Transmit) Status")
    MODEM_STATUS = (0x8A, "Modem Status")
    TRANSMIT_STATUS = (0x8B, "Transmit Status")
    IO_DATA_SAMPLE_RX_INDICATOR_WIFI = (0x8F, "IO Data Sample RX Indicator (Wi-Fi)")
    RECEIVE_PACKET = (0x90, "Receive Packet")
    EXPLICIT_RX_INDICATOR = (0x91, "Explicit RX Indicator")
    IO_DATA_SAMPLE_RX_INDICATOR = (0x92, "IO Data Sample RX Indicator")
    REMOTE_AT_COMMAND_RESPONSE = (0x97, "Remote Command Response")
    RX_SMS = (0x9F, "RX SMS")
    RX_IPV4 = (0xB0, "RX IPv4")
    SEND_DATA_RESPONSE = (0xB8, "Send Data Response")
    DEVICE_REQUEST = (0xB9, "Device Request")
    DEVICE_RESPONSE_STATUS = (0xBA, "Device Response Status")
    FRAME_ERROR = (0xFE, "Frame Error")
    GENERIC = (0xFF, "Generic")
    UNKNOWN = (-1, "Unknown Packet")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the ApiFrameType element.

        Returns:
            Integer: the code of the ApiFrameType element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the ApiFrameType element.

        Returns:
            Integer: the description of the ApiFrameType element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Retrieves the api frame type associated to the given ID.

        Args:
            code (Integer): the code of the API frame type to get.

        Returns:
            :class:`.ApiFrameType`: the API frame type associated to the given code or ``UNKNOWN`` if
                the given code is not a valid ApiFrameType code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return ApiFrameType.UNKNOWN

    code = property(__get_code)
    """Integer. The API frame type code."""

    description = property(__get_description)
    """String. The API frame type description."""


# Dictionary<Integer, ApiFrameType: Used to determine the ApiFrameType from Integer.
ApiFrameType.lookupTable = {x.code: x for x in ApiFrameType}
ApiFrameType.__doc__ += utils.doc_enum(ApiFrameType)



# Copyright 2017, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import re


class XBeeMessage(object):
    """
    This class represents a XBee message, which is formed by a :class:`.RemoteXBeeDevice`
    (the sender) and some data (the data sent) as a bytearray.
    """

    def __init__(self, data, remote_device, timestamp, broadcast=False):
        """
        Class  constructor.

        Args:
            data (Bytearray): the data sent.
            remote_device (:class:`.RemoteXBeeDevice`): the sender.
            broadcast (Boolean, optional, default=``False``): flag indicating whether the  message is
                broadcast (``True``) or not (``False``). Optional.
            timestamp: instant of time when the message was received.
        """
        self.__data = data
        self.__remote_device = remote_device
        self.__is_broadcast = broadcast
        self.__timestamp = timestamp

    def __get_data(self):
        """
        Returns the data of the message.

        Returns:
            Bytearray: the data of the message.
        """
        return self.__data

    def __get_remote_device(self):
        """
        Returns the device which has sent the message.

        Returns:
            :class:`.RemoteXBeeDevice`: the device which has sent the message.
        """
        return self.__remote_device

    def __is_broadcast(self):
        """
        Returns whether the message is broadcast or not.

        Returns:
            Boolean: ``True`` if the message is broadcast, ``False`` otherwise.
        """
        return self.__is_broadcast

    def __get_timestamp(self):
        """
        Returns the moment when the message was received as a ``time.time()``
        function returned value.

        Returns:
            Float: the returned value of using :meth:`time.time()` function when the message was received.
        """
        return self.__timestamp

    def to_dict(self):
        """
        Returns the message information as a dictionary.
        """
        return {"Data: ":        self.__data,
                "Sender: ":      str(self.__remote_device.get_64bit_addr()),
                "Broadcast: ":   self.__is_broadcast,
                "Received at: ": self.__timestamp}

    data = property(__get_data)
    """Bytearray. Bytearray containing the data of the message."""

    remote_device = property(__get_remote_device)
    """:class:`.RemoteXBeeDevice`. The device that has sent the message."""

    is_broadcast = property(__is_broadcast)
    """Boolean. ``True`` to indicate that the message is broadcast, ``False`` otherwise."""
    
    timestamp = property(__get_timestamp)
    """Integer. Instant of time when the message was received."""


class ExplicitXBeeMessage(XBeeMessage):
    """
    This class represents an Explicit XBee message, which is formed by all parameters of a common XBee message and:
    Source endpoint, destination endpoint, cluster ID, profile ID.
    """

    def __init__(self, data, remote_device, timestamp, source_endpoint,
                 dest_endpoint, cluster_id, profile_id, broadcast=False):
        """
        Class constructor.

        Args:
            data (Bytearray): the data sent.
            remote_device (:class:`.RemoteXBeeDevice`): the sender device.
            timestamp: instant of time when the message was received.
            source_endpoint (Integer): source endpoint of the message. 1 byte.
            dest_endpoint (Integer): destination endpoint of the message. 1 byte.
            cluster_id (Integer): cluster id of the message. 2 bytes.
            profile_id (Integer): profile id of the message. 2 bytes.
            broadcast (Boolean, optional, default=``False``): flag indicating whether the message is
                broadcast (``True``) or not (``False``). Optional.
        """
        XBeeMessage.__init__(self, data, remote_device, timestamp, broadcast)
        self.__source_endpoint = source_endpoint
        self.__dest_endpoint = dest_endpoint
        self.__cluster_id = cluster_id
        self.__profile_id = profile_id

    def __get_source_endpoint(self):
        """
        Returns the source endpoint of the message.

        Returns:
            Integer: the source endpoint of the message. 1 byte.
        """
        return self.__source_endpoint

    def __get_dest_endpoint(self):
        """
        Returns the destination endpoint of the message.

        Returns:
            Integer: the destination endpoint of the message. 1 byte.
        """
        return self.__dest_endpoint

    def __get_cluster_id(self):
        """
        Returns the cluster ID of the message.

        Returns:
            Integer: the cluster ID of the message. 2 bytes.
        """
        return self.__cluster_id

    def __get_profile_id(self):
        """
        Returns the profile ID of the message.

        Returns:
            Integer: the profile ID of the message. 2 bytes.
        """
        return self.__profile_id

    def __set_source_endpoint(self, source_endpoint):
        """
        Sets the source endpoint of the message.

        Args:
            source_endpoint (Integer): the new source endpoint of the message.
        """
        self.__source_endpoint = source_endpoint

    def __set_dest_endpoint(self, dest_endpoint):
        """
         Sets the destination endpoint of the message.

         Args:
             dest_endpoint (Integer): the new destination endpoint of the message.
         """
        self.__dest_endpoint = dest_endpoint

    def __set_cluster_id(self, cluster_id):
        """
         Sets the cluster ID of the message.

         Args:
             cluster_id (Integer): the new cluster ID of the message.
         """
        self.__cluster_id = cluster_id

    def __set_profile_id(self, profile_id):
        """
         Sets the profile ID of the message.

         Args:
             profile_id (Integer): the new profile ID of the message.
         """
        self.__profile_id = profile_id

    def to_dict(self):
        dc = XBeeMessage.to_dict(self)
        dc.update({"Src_endpoint":  self.__source_endpoint,
                   "Dest_endpoint": self.__dest_endpoint,
                   "Cluster_id":    self.__cluster_id,
                   "Profile_id":    self.__profile_id})
        return dc

    source_endpoint = property(__get_source_endpoint, __set_source_endpoint)
    """Integer. The source endpoint of the message"""

    dest_endpoint = property(__get_dest_endpoint, __set_dest_endpoint)
    """Integer. The destination endpoint of the message"""

    cluster_id = property(__get_cluster_id, __set_cluster_id)
    """Integer. The Cluster ID of the message."""

    profile_id = property(__get_profile_id, __set_profile_id)
    """Integer. The profile ID of the message."""


class IPMessage(object):
    """
    This class represents an IP message containing the IP address the message belongs to, the source and destination
    ports, the IP protocol, and the content (data) of the message.
    """

    def __init__(self, ip_addr, source_port, dest_port, protocol, data):
        """
        Class  constructor.

        Args:
            ip_addr (:class:`ipaddress.IPv4Address`): The IP address the message comes from.
            source_port (Integer): TCP or UDP source port of the transmission.
            dest_port (Integer): TCP or UDP destination port of the transmission.
            protocol (:class:`.IPProtocol`): IP protocol used in the transmission.
            data (Bytearray): the data sent.

        Raises:
            ValueError: if ``ip_addr`` is ``None``.
            ValueError: if ``protocol`` is ``None``.
            ValueError: if ``data`` is ``None``.
            ValueError: if ``source_port`` is less than 0 or greater than 65535.
            ValueError: if ``dest_port`` is less than 0 or greater than 65535.
        """
        if ip_addr is None:
            raise ValueError("IP address cannot be None")
        if protocol is None:
            raise ValueError("Protocol cannot be None")
        if data is None:
            raise ValueError("Data cannot be None")

        if not 0 <= source_port <= 65535:
            raise ValueError("Source port must be between 0 and 65535")
        if not 0 <= dest_port <= 65535:
            raise ValueError("Destination port must be between 0 and 65535")

        self.__ip_addr = ip_addr
        self.__source_port = source_port
        self.__dest_port = dest_port
        self.__protocol = protocol
        self.__data = data

    def __get_ip_addr(self):
        """
        Returns the IPv4 address this message is associated to.

        Returns:
            :class:`ipaddress.IPv4Address`: The IPv4 address this message is associated to.
        """
        return self.__ip_addr

    def __get_source_port(self):
        """
        Returns the source port of the transmission.

        Returns:
            Integer: The source port of the transmission.
        """
        return self.__source_port

    def __get_dest_port(self):
        """
        Returns the destination port of the transmission.

        Returns:
            Integer: The destination port of the transmission.
        """
        return self.__dest_port

    def __get_protocol(self):
        """
        Returns the protocol used in the transmission.

        Returns:
            :class:`.IPProtocol`: The protocol used in the transmission.
        """
        return self.__protocol

    def __get_data(self):
        """
        Returns the data of the message.

        Returns:
            Bytearray: the data of the message.
        """
        return self.__data

    def to_dict(self):
        """
        Returns the message information as a dictionary.
        """
        return {"IP address: ":       self.__ip_addr,
                "Source port: ":      self.__source_port,
                "Destination port: ": self.__dest_port,
                "Protocol: ":         self.__protocol,
                "Data: ":             self.__data}

    ip_addr = property(__get_ip_addr)
    """:class:`ipaddress.IPv4Address`. The IPv4 address this message is associated to."""

    source_port = property(__get_source_port)
    """Integer. The source port of the transmission."""

    dest_port = property(__get_dest_port)
    """Integer. The destination port of the transmission."""

    protocol = property(__get_protocol)
    """:class:`.IPProtocol`. The protocol used in the transmission."""

    data = property(__get_data)
    """Bytearray. Bytearray containing the data of the message."""


class SMSMessage(object):
    """
    This class represents an SMS message containing the phone number that sent
    the message and the content (data) of the message.

    This class is used within the library to read SMS sent to Cellular devices.
    """

    __PHONE_NUMBER_PATTERN = "^\+?\d+$"

    def __init__(self, phone_number, data):
        """
        Class  constructor. Instantiates a new :class:`.SMSMessage` object with the provided parameters.

        Args:
            phone_number (String): The phone number that sent the message.
            data (String): The message text.

        Raises:
            ValueError: if ``phone_number`` is ``None``.
            ValueError: if ``data`` is ``None``.
            ValueError: if ``phone_number`` is not a valid phone number.
        """
        if phone_number is None:
            raise ValueError("Phone number cannot be None")
        if data is None:
            raise ValueError("Data cannot be None")
        if not re.compile(SMSMessage.__PHONE_NUMBER_PATTERN).match(phone_number):
            raise ValueError("Invalid phone number")

        self.__phone_number = phone_number
        self.__data = data

    def __get_phone_number(self):
        """
        Returns the phone number that sent the message.

        Returns:
            String: The phone number that sent the message.
        """
        return self.__phone_number

    def __get_data(self):
        """
        Returns the data of the message.

        Returns:
            String: The data of the message.
        """
        return self.__data

    def to_dict(self):
        """
        Returns the message information as a dictionary.
        """
        return {"Phone number: ": self.__phone_number,
                "Data: ":         self.__data}

    phone_number = property(__get_phone_number)
    """String. The phone number that sent the message."""

    data = property(__get_data)
    """String. The data of the message."""



# Copyright 2017, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from serial import Serial, EIGHTBITS, STOPBITS_ONE, PARITY_NONE
import enum
import digi.xbee.exception


class FlowControl(enum.Enum):
    """
    This class represents all available flow controls.
    """

    NONE = None
    SOFTWARE = 0
    HARDWARE_RTS_CTS = 1
    HARDWARE_DSR_DTR = 2
    UNKNOWN = 99


class XBeeSerialPort(Serial):
    """
    This class extends the functionality of Serial class (PySerial).

    .. seealso::
       | _PySerial: https://github.com/pyserial/pyserial
    """

    __DEFAULT_PORT_TIMEOUT = 0.1  # seconds
    __DEFAULT_DATA_BITS = EIGHTBITS
    __DEFAULT_STOP_BITS = STOPBITS_ONE
    __DEFAULT_PARITY = PARITY_NONE
    __DEFAULT_FLOW_CONTROL = FlowControl.NONE

    def __init__(self, baud_rate, port,
                 data_bits=__DEFAULT_DATA_BITS, stop_bits=__DEFAULT_STOP_BITS, parity=__DEFAULT_PARITY,
                 flow_control=__DEFAULT_FLOW_CONTROL, timeout=__DEFAULT_PORT_TIMEOUT):
        """
        Class constructor. Instantiates a new ``XBeeSerialPort`` object with the given
        port parameters.

        Args:
            baud_rate (Integer): serial port baud rate.
            port (String): serial port name to use.
            data_bits (Integer, optional): serial data bits. Default to 8.
            stop_bits (Float, optional): serial stop bits. Default to 1.
            parity (Char, optional): serial parity. Default to 'N' (None).
            flow_control (Integer, optional): serial flow control. Default to ``None``.
            timeout (Integer, optional): read timeout. Default to 0.1 seconds.

        .. seealso::
           | _PySerial: https://github.com/pyserial/pyserial
        """
        if flow_control == FlowControl.SOFTWARE:
            Serial.__init__(self, port=port, baudrate=baud_rate,
                            bytesize=data_bits, stopbits=stop_bits, parity=parity, timeout=timeout, xonxoff=True)
        elif flow_control == FlowControl.HARDWARE_DSR_DTR:
            Serial.__init__(self, port=port, baudrate=baud_rate,
                            bytesize=data_bits, stopbits=stop_bits, parity=parity, timeout=timeout, dsrdtr=True)
        elif flow_control == FlowControl.HARDWARE_RTS_CTS:
            Serial.__init__(self, port=port, baudrate=baud_rate,
                            bytesize=data_bits, stopbits=stop_bits, parity=parity, timeout=timeout, rtscts=True)
        else:
            Serial.__init__(self, port=port, baudrate=baud_rate,
                            bytesize=data_bits, stopbits=stop_bits, parity=parity, timeout=timeout)
        self._isOpen = True if port is not None else False

    def read_byte(self):
        """
        Synchronous. Reads one byte from serial port.

        Returns:
            Integer: the read byte.

        Raises:
            TimeoutException: if there is no bytes ins serial port buffer.
        """
        byte = bytearray(self.read(1))
        if len(byte) == 0:
            raise digi.xbee.exception.TimeoutException()
        else:
            return byte[0]

    def read_bytes(self, num_bytes):
        """
        Synchronous. Reads the specified number of bytes from the serial port.

        Args:
            num_bytes (Integer): the number of bytes to read.

        Returns:
            Bytearray: the read bytes.

        Raises:
            TimeoutException: if the number of bytes read is less than ``num_bytes``.
        """
        read_bytes = bytearray(self.read(num_bytes))
        if len(read_bytes) != num_bytes:
            raise digi.xbee.exception.TimeoutException()
        return read_bytes

    def read_existing(self):
        """
        Asynchronous. Reads all bytes in the serial port buffer. May read 0 bytes.

        Returns:
            Bytearray: the bytes read.
        """
        return bytearray(self.read(self.inWaiting()))

    def get_read_timeout(self):
        """
        Returns the serial port read timeout.

        Returns:
            Integer: read timeout in seconds.
        """
        return self.timeout

    def set_read_timeout(self, read_timeout):
        """
        Sets the serial port read timeout in seconds.

        Args:
            read_timeout (Integer): the new serial port read timeout in seconds.
        """
        self.timeout = read_timeout



# Copyright 2017, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


class XBeeException(Exception):
    """
    Generic XBee API exception. This class and its subclasses indicate
    conditions that an application might want to catch.

    All functionality of this class is the inherited of `Exception
    <https://docs.python.org/2/library/exceptions.html?highlight=exceptions.exception#exceptions.Exception>`_.
    """
    pass


class CommunicationException(XBeeException):
    """
    This exception will be thrown when any problem related to the communication 
    with the XBee device occurs.

    All functionality of this class is the inherited of `Exception
    <https://docs.python.org/2/library/exceptions.html?highlight=exceptions.exception#exceptions.Exception>`_.
    """
    pass


class ATCommandException(CommunicationException):
    """
    This exception will be thrown when a response of a packet is not success or OK.

    All functionality of this class is the inherited of `Exception
    <https://docs.python.org/2/library/exceptions.html?highlight=exceptions.exception#exceptions.Exception>`_.
    """
    pass


class ConnectionException(XBeeException):
    """
    This exception will be thrown when any problem related to the connection 
    with the XBee device occurs.

    All functionality of this class is the inherited of `Exception
    <https://docs.python.org/2/library/exceptions.html?highlight=exceptions.exception#exceptions.Exception>`_.
    """
    pass


class XBeeDeviceException(XBeeException):
    """
    This exception will be thrown when any problem related to the XBee device 
    occurs.

    All functionality of this class is the inherited of `Exception
    <https://docs.python.org/2/library/exceptions.html?highlight=exceptions.exception#exceptions.Exception>`_.
    """
    pass


class InvalidConfigurationException(ConnectionException):
    """
    This exception will be thrown when trying to open an interface with an 
    invalid configuration.

    All functionality of this class is the inherited of `Exception
    <https://docs.python.org/2/library/exceptions.html?highlight=exceptions.exception#exceptions.Exception>`_.
    """
    __DEFAULT_MESSAGE = "The configuration used to open the interface is invalid."

    def __init__(self, message=__DEFAULT_MESSAGE):
        ConnectionException.__init__(self, message)


class InvalidOperatingModeException(ConnectionException):
    """
    This exception will be thrown if the operating mode is different than 
    *OperatingMode.API_MODE* and *OperatingMode.API_MODE*

    All functionality of this class is the inherited of `Exception
    <https://docs.python.org/2/library/exceptions.html?highlight=exceptions.exception#exceptions.Exception>`_.
    """
    __DEFAULT_MESSAGE = "The operating mode of the XBee device is not supported by the library."

    def __init__(self, message=__DEFAULT_MESSAGE):
        ConnectionException.__init__(self, message)

    @classmethod
    def from_operating_mode(cls, operating_mode):
        """
        Class constructor.

        Args:
            operating_mode (:class:`.OperatingMode`): the operating mode that generates the exceptions.
        """
        return cls("Unsupported operating mode: " + operating_mode.description)


class InvalidPacketException(CommunicationException):
    """
    This exception will be thrown when there is an error parsing an API packet 
    from the input stream.

    All functionality of this class is the inherited of `Exception
    <https://docs.python.org/2/library/exceptions.html?highlight=exceptions.exception#exceptions.Exception>`_.
    """
    __DEFAULT_MESSAGE = "The XBee API packet is not properly formed."

    def __init__(self, message=__DEFAULT_MESSAGE):
        CommunicationException.__init__(self, message)


class OperationNotSupportedException(XBeeDeviceException):
    """
    This exception will be thrown when the operation performed is not supported 
    by the XBee device.

    All functionality of this class is the inherited of `Exception
    <https://docs.python.org/2/library/exceptions.html?highlight=exceptions.exception#exceptions.Exception>`_.
    """
    __DEFAULT_MESSAGE = "The requested operation is not supported by either the connection interface or " \
                        "the XBee device."

    def __init__(self, message=__DEFAULT_MESSAGE):
        XBeeDeviceException.__init__(self, message)


class TimeoutException(CommunicationException):
    """
    This exception will be thrown when performing synchronous operations and 
    the configured time expires.

    All functionality of this class is the inherited of `Exception
    <https://docs.python.org/2/library/exceptions.html?highlight=exceptions.exception#exceptions.Exception>`_.
    """
    __DEFAULT_MESSAGE = "There was a timeout while executing the requested operation."

    def __init__(self, _message=__DEFAULT_MESSAGE):
        CommunicationException.__init__(self)


class TransmitException(CommunicationException):
    """
    This exception will be thrown when receiving a transmit status different 
    than *TransmitStatus.SUCCESS* after sending an XBee API packet.

    All functionality of this class is the inherited of `Exception
    <https://docs.python.org/2/library/exceptions.html?highlight=exceptions.exception#exceptions.Exception>`_.
    """
    __DEFAULT_MESSAGE = "There was a problem with a transmitted packet response (status not ok)"

    def __init__(self, _message=__DEFAULT_MESSAGE):
        CommunicationException.__init__(self, _message)



# Copyright 2017, Digi International Inc.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from digi.xbee.devices import XBeeDevice

# TODO: Replace with the serial port where your local module is connected to. 
PORT = "COM1"
# TODO: Replace with the baud rate of your local module.
BAUD_RATE = 9600


def main():
    print(" +-------------------------------------------------+")
    print(" | XBee Python Library Receive Data Polling Sample |")
    print(" +-------------------------------------------------+\n")

    device = XBeeDevice(PORT, BAUD_RATE)

    try:
        device.open()

        device.flush_queues()

        print("Waiting for data...\n")

        while True:
            xbee_message = device.read_data()
            if xbee_message is not None:
                print("From %s >> %s" % (xbee_message.remote_device.get_64bit_addr(),
                                         xbee_message.data.decode()))

    finally:
        if device is not None and device.is_open():
            device.close()


if __name__ == '__main__':
    main()



# Copyright 2017, Digi International Inc.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from digi.xbee.devices import XBeeDevice
from digi.xbee.io import IOLine, IOMode
import time
import threading

# TODO: Replace with the serial port where your local module is connected to. 
PORT = "COM1"
# TODO: Replace with the baud rate of your local module.
BAUD_RATE = 9600

REMOTE_NODE_ID = "REMOTE"

IOLINE_IN = IOLine.DIO1_AD1


def main():
    print(" +--------------------------------------------+")
    print(" | XBee Python Library Read Remote ADC Sample |")
    print(" +--------------------------------------------+\n")

    stop = False
    th = None

    local_device = XBeeDevice(PORT, BAUD_RATE)

    try:
        local_device.open()

        # Obtain the remote XBee device from the XBee network.
        xbee_network = local_device.get_network()
        remote_device = xbee_network.discover_device(REMOTE_NODE_ID)
        if remote_device is None:
            print("Could not find the remote device")
            exit(1)

        remote_device.set_io_configuration(IOLINE_IN, IOMode.ADC)

        def read_adc_task():
            while not stop:
                # Read the analog value from the remote input line.
                value = remote_device.get_adc_value(IOLINE_IN)
                print("%s: %d" % (IOLINE_IN, value))
                
                time.sleep(0.2)

        th = threading.Thread(target=read_adc_task)

        time.sleep(0.5)
        th.start()

        input()

    finally:
        stop = True
        if th is not None and th.isAlive():
            th.join()
        if local_device is not None and local_device.is_open():
            local_device.close()


if __name__ == '__main__':
    main()



# Copyright 2017, 2018, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from concurrent.futures import ThreadPoolExecutor
from queue import Queue, Empty
import logging
import threading
import time

import digi.xbee.devices
from digi.xbee.models.atcomm import SpecialByte
from digi.xbee.models.address import XBee64BitAddress, XBee16BitAddress
from digi.xbee.models.message import XBeeMessage, ExplicitXBeeMessage, IPMessage, \
    SMSMessage
from digi.xbee.models.mode import OperatingMode
from digi.xbee.models.options import ReceiveOptions
from digi.xbee.models.protocol import XBeeProtocol
from digi.xbee.packets import factory
from digi.xbee.packets.aft import ApiFrameType
from digi.xbee.packets.base import XBeePacket, XBeeAPIPacket
from digi.xbee.packets.common import ReceivePacket
from digi.xbee.packets.raw import RX64Packet, RX16Packet
from digi.xbee.util import utils
from digi.xbee.exception import TimeoutException, InvalidPacketException
from digi.xbee.io import IOSample


# Maximum number of parallel callbacks.
MAX_PARALLEL_CALLBACKS = 50

executor = ThreadPoolExecutor(max_workers=MAX_PARALLEL_CALLBACKS)


class XBeeEvent(list):
    """
    This class represents a generic XBee event.

    New event callbacks can be added here following this prototype:

    ::

        def callback_prototype(*args, **kwargs):
            #do something...

    All of them will be executed when the event is fired.

    .. seealso::
       | list (Python standard class)
    """
    def __call__(self, *args, **kwargs):
        for f in self:
            future = executor.submit(f, *args, **kwargs)
            future.add_done_callback(self.__execution_finished)

    def __repr__(self):
        return "Event(%s)" % list.__repr__(self)

    def __iadd__(self, other):
        self.append(other)
        return self

    def __isub__(self, other):
        self.remove(other)
        return self

    def __execution_finished(self, future):
        """
        Called when the execution of the callable has finished.

        Args:
            future (:class:`.Future`): Future associated to the execution of the callable.

        Raises:
            Exception: if the execution of the callable raised any exception.
        """
        if future.exception():
            raise future.exception()


class PacketReceived(XBeeEvent):
    """
    This event is fired when an XBee receives any packet, independent of
    its frame type.

    The callbacks for handle this events will receive the following arguments:
        1. received_packet (:class:`.XBeeAPIPacket`): the received packet.

    .. seealso::
       | :class:`.XBeeAPIPacket`
       | :class:`.XBeeEvent`
    """
    pass


class DataReceived(XBeeEvent):
    """
    This event is fired when an XBee receives data.

    The callbacks for handle this events will receive the following arguments:
        1. message (:class:`.XBeeMessage`): message containing the data received, the sender and the time.

    .. seealso::
       | :class:`.XBeeEvent`
       | :class:`.XBeeMessage`
    """
    pass


class ModemStatusReceived(XBeeEvent):
    """
    This event is fired when a XBee receives a modem status packet.

    The callbacks for handle this events will receive the following arguments:
        1. modem_status (:class:`.ModemStatus`): the modem status received.

    .. seealso::
       | :class:`.XBeeEvent`
       | :class:`.ModemStatus`
    """
    pass


class IOSampleReceived(XBeeEvent):
    """
    This event is fired when a XBee receives an IO packet.

    This includes:

    1. IO data sample RX indicator packet.
    2. RX IO 16 packet.
    3. RX IO 64 packet.

    The callbacks that handle this event will receive the following arguments:
        1. io_sample (:class:`.IOSample`): the received IO sample.
        2. sender (:class:`.RemoteXBeeDevice`): the remote XBee device who has sent the packet.
        3. time (Integer): the time in which the packet was received.

    .. seealso::
       | :class:`.IOSample`
       | :class:`.RemoteXBeeDevice`
       | :class:`.XBeeEvent`
    """
    pass


class DeviceDiscovered(XBeeEvent):
    """
    This event is fired when an XBee discovers another remote XBee
    during a discovering operation.

    The callbacks that handle this event will receive the following arguments:
        1. discovered_device (:class:`.RemoteXBeeDevice`): the discovered remote XBee device.

    .. seealso::
       | :class:`.RemoteXBeeDevice`
       | :class:`.XBeeEvent`
    """
    pass


class DiscoveryProcessFinished(XBeeEvent):
    """
    This event is fired when the discovery process finishes, either
    successfully or due to an error.

    The callbacks that handle this event will receive the following arguments:
        1. status (:class:`.NetworkDiscoveryStatus`): the network discovery status.

    .. seealso::
       | :class:`.NetworkDiscoveryStatus`
       | :class:`.XBeeEvent`
    """
    pass


class ExplicitDataReceived(XBeeEvent):
    """
    This event is fired when an XBee receives an explicit data packet.

    The callbacks for handle this events will receive the following arguments:
        1. message (:class:`.ExplicitXBeeMessage`): message containing the data received, the sender, the time
            and explicit data message parameters.

    .. seealso::
       | :class:`.XBeeEvent`
       | :class:`.XBeeMessage`
    """
    pass


class IPDataReceived(XBeeEvent):
    """
    This event is fired when an XBee receives IP data.

    The callbacks for handle this events will receive the following arguments:
        1. message (:class:`.IPMessage`): message containing containing the IP address the message
            belongs to, the source and destination ports, the IP protocol, and the content (data) of the message.

    .. seealso::
       | :class:`.XBeeEvent`
       | :class:`.IPMessage`
    """
    pass


class SMSReceived(XBeeEvent):
    """
    This event is fired when an XBee receives an SMS.

    The callbacks for handle this events will receive the following arguments:
        1. message (:class:`.SMSMessage`): message containing the phone number that sent
            the message and the content (data) of the message.

    .. seealso::
       | :class:`.XBeeEvent`
       | :class:`.SMSMessage`
    """
    pass


class PacketListener(threading.Thread):
    """
    This class represents a packet listener, which is a thread that's always
    listening for incoming packets to the XBee.

    When it receives a packet, this class throws an event depending on which
    packet it is. You can add your own callbacks for this events via certain
    class methods. This callbacks must have a certain header, see each event 
    documentation.

    This class has fields that are events. Its recommended to use only the
    append() and remove() method on them, or -= and += operators.
    If you do something more with them, it's for your own risk.

    Here are the parameters which will be received by the event callbacks, 
    depending on which event it is in each case:

    The following parameters are passed via \*\*kwargs to event callbacks of:

    1. PacketReceived:
        1.1 received_packet (:class:`.XBeeAPIPacket`): the received packet.
        1.2 sender (:class:`.RemoteXBeeDevice`): the remote XBee device who has sent the packet.
    2. DataReceived
        2.1 message (:class:`.XBeeMessage`): message containing the data received, the sender and the time.
    3. ModemStatusReceived
        3.1 modem_status (:class:`.ModemStatus`): the modem status received.
    """

    __DEFAULT_QUEUE_MAX_SIZE = 40
    """
    Default max. size that the queue has.
    """

    _LOG_PATTERN = "{port:<6s}{event:<12s}{fr_type:<10s}{sender:<18s}{more_data:<50s}"
    """
    Generic pattern for display received messages (high-level) with logger.
    """

    _log = logging.getLogger(__name__)
    """
    Logger.
    """

    def __init__(self, serial_port, xbee_device, queue_max_size=None):
        """
        Class constructor. Instantiates a new :class:`.PacketListener` object with the provided parameters.

        Args:
            serial_port (:class:`.XbeeSerialPort`): the COM port to which this listener will be listening.
            xbee_device (:class:`.XBeeDevice`): the XBee that is the listener owner.
            queue_max_size (Integer): the maximum size of the XBee queue.
        """
        threading.Thread.__init__(self)

        # User callbacks:
        self.__packet_received = PacketReceived()
        self.__data_received = DataReceived()
        self.__modem_status_received = ModemStatusReceived()
        self.__io_sample_received = IOSampleReceived()
        self.__explicit_packet_received = ExplicitDataReceived()
        self.__ip_data_received = IPDataReceived()
        self.__sms_received = SMSReceived()

        # API internal callbacks:
        self.__packet_received_API = xbee_device.get_xbee_device_callbacks()

        self.__xbee_device = xbee_device
        self.__serial_port = serial_port
        self.__stop = True

        self.__queue_max_size = queue_max_size if queue_max_size is not None else self.__DEFAULT_QUEUE_MAX_SIZE
        self.__xbee_queue = XBeeQueue(self.__queue_max_size)
        self.__data_xbee_queue = XBeeQueue(self.__queue_max_size)
        self.__explicit_xbee_queue = XBeeQueue(self.__queue_max_size)
        self.__ip_xbee_queue = XBeeQueue(self.__queue_max_size)

        self._log_handler = logging.StreamHandler()
        self._log.addHandler(self._log_handler)

    def __del__(self):
        self._log.removeHandler(self._log_handler)

    def run(self):
        """
        This is the method that will be executing for listening packets.

        For each packet, it will execute the proper callbacks.
        """
        try:
            self.__stop = False
            while not self.__stop:
                # Try to read a packet. Read packet is unescaped.
                raw_packet = self.__try_read_packet(self.__xbee_device.operating_mode)

                if raw_packet is not None:
                    # If the current protocol is 802.15.4, the packet may have to be discarded.
                    if (self.__xbee_device.get_protocol() == XBeeProtocol.RAW_802_15_4 and
                       not self.__check_packet_802_15_4(raw_packet)):
                        continue

                    # Build the packet.
                    try:
                        read_packet = factory.build_frame(raw_packet, self.__xbee_device.operating_mode)
                    except InvalidPacketException as e:
                        self._log.error("Error processing packet '%s': %s" % (utils.hex_to_string(raw_packet), str(e)))
                        continue

                    self._log.debug(self.__xbee_device.LOG_PATTERN.format(port=self.__xbee_device.serial_port.port,
                                                                          event="RECEIVED",
                                                                          opmode=self.__xbee_device.operating_mode,
                                                                          content=utils.hex_to_string(raw_packet)))

                    # Add the packet to the queue.
                    self.__add_packet_queue(read_packet)

                    # If the packet has information about a remote device, extract it
                    # and add/update this remote device to/in this XBee's network.
                    remote = self.__try_add_remote_device(read_packet)

                    # Execute API internal callbacks.
                    self.__packet_received_API(read_packet)

                    # Execute all user callbacks.
                    self.__execute_user_callbacks(read_packet, remote)
        except Exception as e:
            if not self.__stop:
                self._log.exception(e)
        finally:
            if not self.__stop:
                self.__stop = True
                if self.__serial_port.isOpen():
                    self.__serial_port.close()

    def stop(self):
        """
        Stops listening.
        """
        self.__stop = True

    def is_running(self):
        """
        Returns whether this instance is running or not.

        Returns:
            Boolean: ``True`` if this instance is running, ``False`` otherwise.
        """
        return not self.__stop

    def get_queue(self):
        """
        Returns the packets queue.

        Returns:
            :class:`.XBeeQueue`: the packets queue.
        """
        return self.__xbee_queue

    def get_data_queue(self):
        """
        Returns the data packets queue.

        Returns:
            :class:`.XBeeQueue`: the data packets queue.
        """
        return self.__data_xbee_queue

    def get_explicit_queue(self):
        """
        Returns the explicit packets queue.

        Returns:
            :class:`.XBeeQueue`: the explicit packets queue.
        """
        return self.__explicit_xbee_queue

    def get_ip_queue(self):
        """
        Returns the IP packets queue.

        Returns:
            :class:`.XBeeQueue`: the IP packets queue.
        """
        return self.__ip_xbee_queue

    def add_packet_received_callback(self, callback):
        """
        Adds a callback for the event :class:`.PacketReceived`.

        Args:
            callback (Function): the callback. Receives two arguments.

                * The received packet as a :class:`.XBeeAPIPacket`
                * The sender as a :class:`.RemoteXBeeDevice`
        """
        self.__packet_received += callback

    def add_data_received_callback(self, callback):
        """
        Adds a callback for the event :class:`.DataReceived`.

        Args:
            callback (Function): the callback. Receives one argument.

                * The data received as an :class:`.XBeeMessage`
        """
        self.__data_received += callback

    def add_modem_status_received_callback(self, callback):
        """
        Adds a callback for the event :class:`.ModemStatusReceived`.

        Args:
            callback (Function): the callback. Receives one argument.

                * The modem status as a :class:`.ModemStatus`
        """
        self.__modem_status_received += callback

    def add_io_sample_received_callback(self, callback):
        """
        Adds a callback for the event :class:`.IOSampleReceived`.

        Args:
            callback (Function): the callback. Receives three arguments.

                * The received IO sample as an :class:`.IOSample`
                * The remote XBee device who has sent the packet as a :class:`.RemoteXBeeDevice`
                * The time in which the packet was received as an Integer
        """
        self.__io_sample_received += callback

    def add_explicit_data_received_callback(self, callback):
        """
        Adds a callback for the event :class:`.ExplicitDataReceived`.

        Args:
            callback (Function): the callback. Receives one argument.

                * The explicit data received as an :class:`.ExplicitXBeeMessage`
        """
        self.__explicit_packet_received += callback

    def add_ip_data_received_callback(self, callback):
        """
        Adds a callback for the event :class:`.IPDataReceived`.

        Args:
            callback (Function): the callback. Receives one argument.

                * The data received as an :class:`.IPMessage`
        """
        self.__ip_data_received += callback

    def add_sms_received_callback(self, callback):
        """
        Adds a callback for the event :class:`.SMSReceived`.

        Args:
            callback (Function): the callback. Receives one argument.

                * The data received as an :class:`.SMSMessage`
        """
        self.__sms_received += callback

    def del_packet_received_callback(self, callback):
        """
        Deletes a callback for the callback list of :class:`.PacketReceived` event.

        Args:
            callback (Function): the callback to delete.

        Raises:
            ValueError: if ``callback`` is not in the callback list of :class:`.PacketReceived` event.
        """
        self.__packet_received -= callback

    def del_data_received_callback(self, callback):
        """
        Deletes a callback for the callback list of :class:`.DataReceived` event.

        Args:
            callback (Function): the callback to delete.

        Raises:
            ValueError: if ``callback`` is not in the callback list of :class:`.DataReceived` event.
        """
        self.__data_received -= callback

    def del_modem_status_received_callback(self, callback):
        """
        Deletes a callback for the callback list of :class:`.ModemStatusReceived` event.

        Args:
            callback (Function): the callback to delete.

        Raises:
            ValueError: if ``callback`` is not in the callback list of :class:`.ModemStatusReceived` event.
        """
        self.__modem_status_received -= callback

    def del_io_sample_received_callback(self, callback):
        """
        Deletes a callback for the callback list of :class:`.IOSampleReceived` event.

        Args:
            callback (Function): the callback to delete.

        Raises:
            ValueError: if ``callback`` is not in the callback list of :class:`.IOSampleReceived` event.
        """
        self.__io_sample_received -= callback

    def del_explicit_data_received_callback(self, callback):
        """
        Deletes a callback for the callback list of :class:`.ExplicitDataReceived` event.

        Args:
            callback (Function): the callback to delete.

        Raises:
            ValueError: if ``callback`` is not in the callback list of :class:`.ExplicitDataReceived` event.
        """
        self.__explicit_packet_received -= callback

    def del_ip_data_received_callback(self, callback):
        """
        Deletes a callback for the callback list of :class:`.IPDataReceived` event.

        Args:
            callback (Function): the callback to delete.

        Raises:
            ValueError: if ``callback`` is not in the callback list of :class:`.IPDataReceived` event.
        """
        self.__ip_data_received -= callback

    def del_sms_received_callback(self, callback):
        """
        Deletes a callback for the callback list of :class:`.SMSReceived` event.

        Args:
            callback (Function): the callback to delete.

        Raises:
            ValueError: if ``callback`` is not in the callback list of :class:`.SMSReceived` event.
        """
        self.__sms_received -= callback

    def __execute_user_callbacks(self, xbee_packet, remote=None):
        """
        Executes callbacks corresponding to the received packet.

        Args:
            xbee_packet (:class:`.XBeeAPIPacket`): the received packet.
            remote (:class:`.RemoteXBeeDevice`): the XBee device that sent the packet.
        """
        # All packets callback.
        self.__packet_received(xbee_packet)

        # Data reception callbacks
        if (xbee_packet.get_frame_type() == ApiFrameType.RX_64 or
                xbee_packet.get_frame_type() == ApiFrameType.RX_16 or
                xbee_packet.get_frame_type() == ApiFrameType.RECEIVE_PACKET):
            _data = xbee_packet.rf_data
            is_broadcast = xbee_packet.receive_options == ReceiveOptions.BROADCAST_PACKET
            self.__data_received(XBeeMessage(_data, remote, time.time(), is_broadcast))
            self._log.info(self._LOG_PATTERN.format(port=self.__xbee_device.serial_port.port,
                                                    event="RECEIVED",
                                                    fr_type="DATA",
                                                    sender=str(remote.get_64bit_addr()) if remote is not None
                                                    else "None",
                                                    more_data=utils.hex_to_string(xbee_packet.rf_data)))

        # Modem status callbacks
        elif xbee_packet.get_frame_type() == ApiFrameType.MODEM_STATUS:
            self.__modem_status_received(xbee_packet.modem_status)
            self._log.info(self._LOG_PATTERN.format(port=self.__xbee_device.serial_port.port,
                                                    event="RECEIVED",
                                                    fr_type="MODEM STATUS",
                                                    sender=str(remote.get_64bit_addr()) if remote is not None
                                                    else "None",
                                                    more_data=xbee_packet.modem_status))

        # IO_sample callbacks
        elif (xbee_packet.get_frame_type() == ApiFrameType.RX_IO_16 or
              xbee_packet.get_frame_type() == ApiFrameType.RX_IO_64 or
              xbee_packet.get_frame_type() == ApiFrameType.IO_DATA_SAMPLE_RX_INDICATOR):
            self.__io_sample_received(xbee_packet.io_sample, remote, time.time())
            self._log.info(self._LOG_PATTERN.format(port=self.__xbee_device.serial_port.port,
                                                    event="RECEIVED",
                                                    fr_type="IOSAMPLE",
                                                    sender=str(remote.get_64bit_addr()) if remote is not None
                                                    else "None",
                                                    more_data=str(xbee_packet.io_sample)))

        # Explicit packet callbacks
        elif xbee_packet.get_frame_type() == ApiFrameType.EXPLICIT_RX_INDICATOR:
            is_broadcast = False
            # If it's 'special' packet, notify the data_received callbacks too:
            if self.__is_special_explicit_packet(xbee_packet):
                self.__data_received(XBeeMessage(xbee_packet.rf_data, remote, time.time(), is_broadcast))
            self.__explicit_packet_received(PacketListener.__expl_to_message(remote, is_broadcast, xbee_packet))
            self._log.info(self._LOG_PATTERN.format(port=self.__xbee_device.serial_port.port,
                                                    event="RECEIVED",
                                                    fr_type="EXPLICIT DATA",
                                                    sender=str(remote.get_64bit_addr()) if remote is not None
                                                    else "None",
                                                    more_data=utils.hex_to_string(xbee_packet.rf_data)))

        # IP data
        elif xbee_packet.get_frame_type() == ApiFrameType.RX_IPV4:
            self.__ip_data_received(
                IPMessage(xbee_packet.source_address, xbee_packet.source_port,
                          xbee_packet.dest_port, xbee_packet.ip_protocol, xbee_packet.data))
            self._log.info(self._LOG_PATTERN.format(port=self.__xbee_device.serial_port.port,
                                                    event="RECEIVED",
                                                    fr_type="IP DATA",
                                                    sender=str(xbee_packet.source_address),
                                                    more_data=utils.hex_to_string(xbee_packet.data)))

        # SMS
        elif xbee_packet.get_frame_type() == ApiFrameType.RX_SMS:
            self.__sms_received(SMSMessage(xbee_packet.phone_number, xbee_packet.data))
            self._log.info(self._LOG_PATTERN.format(port=self.__xbee_device.serial_port.port,
                                                    event="RECEIVED",
                                                    fr_type="SMS",
                                                    sender=str(xbee_packet.phone_number),
                                                    more_data=xbee_packet.data))

    def __read_next_byte(self, operating_mode):
        """
        Returns the next byte in bytearray format. If the operating mode is
        OperatingMode.ESCAPED_API_MODE, the bytearray could contain 2 bytes.

        If in escaped API mode and the byte that was read was the escape byte,
        it will also read the next byte.

        Args:
            operating_mode (:class:`.OperatingMode`): the operating mode in which the byte should be read.

        Returns:
            Bytearray: the read byte or bytes as bytearray, ``None`` otherwise.
        """
        read_data = bytearray()
        read_byte = self.__serial_port.read_byte()
        read_data.append(read_byte)
        # Read escaped bytes in API escaped mode.
        if operating_mode == OperatingMode.ESCAPED_API_MODE and read_byte == XBeePacket.ESCAPE_BYTE:
            read_data.append(self.__serial_port.read_byte())

        return read_data

    def __try_read_packet(self, operating_mode=OperatingMode.API_MODE):
        """
        Reads the next packet. Starts to read when finds the start delimiter.
        The last byte read is the checksum.

        If there is something in the COM buffer after the
        start delimiter, this method discards it.

        If the method can't read a complete and correct packet,
        it will return ``None``.

        Args:
            operating_mode (:class:`.OperatingMode`): the operating mode in which the packet should be read.

        Returns:
            Bytearray: the read packet as bytearray if a packet is read, ``None`` otherwise.
        """
        try:
            xbee_packet = bytearray(1)
            # Add packet delimiter.
            xbee_packet[0] = self.__serial_port.read_byte()
            while xbee_packet[0] != SpecialByte.HEADER_BYTE.value:
                xbee_packet[0] = self.__serial_port.read_byte()

            # Add packet length.
            packet_length_byte = bytearray()
            for _ in range(0, 2):
                packet_length_byte += self.__read_next_byte(operating_mode)
            xbee_packet += packet_length_byte
            # Length needs to be un-escaped in API escaped mode to obtain its integer equivalent.
            if operating_mode == OperatingMode.ESCAPED_API_MODE:
                length = utils.length_to_int(XBeeAPIPacket.unescape_data(packet_length_byte))
            else:
                length = utils.length_to_int(packet_length_byte)

            # Add packet payload.
            for _ in range(0, length):
                xbee_packet += self.__read_next_byte(operating_mode)

            # Add packet checksum.
            for _ in range(0, 1):
                xbee_packet += self.__read_next_byte(operating_mode)

            # Return the packet unescaped.
            if operating_mode == OperatingMode.ESCAPED_API_MODE:
                return XBeeAPIPacket.unescape_data(xbee_packet)
            else:
                return xbee_packet
        except TimeoutException:
            return None

    def __create_remote_device_from_packet(self, xbee_packet):
        """
        Creates a :class:`.RemoteXBeeDevice` that represents the device that
        has sent the ``xbee_packet``.

        Returns:
            :class:`.RemoteXBeeDevice`
        """
        x64bit_addr, x16bit_addr = self.__get_remote_device_data_from_packet(xbee_packet)
        return digi.xbee.devices.RemoteXBeeDevice(self.__xbee_device, x64bit_addr, x16bit_addr)

    @staticmethod
    def __get_remote_device_data_from_packet(xbee_packet):
        """
        Extracts the 64 bit-address and the 16 bit-address from ``xbee_packet`` if is
        possible.
        """
        x64bit_addr = None
        x16bit_addr = None
        if hasattr(xbee_packet, "x64bit_source_addr"):
            x64bit_addr = xbee_packet.x64bit_source_addr
        if hasattr(xbee_packet, "x16bit_source_addr"):
            x16bit_addr = xbee_packet.x16bit_source_addr
        return x64bit_addr, x16bit_addr

    @staticmethod
    def __check_packet_802_15_4(raw_data):
        """
        If the current XBee's protocol is 802.15.4 and
        the user sends many 'ND' commands, the device could return
        an RX 64 IO packet with an invalid payload (length < 5).

        In this case the packet must be discarded, or an exception
        must be raised.

        This method checks a received raw_data and returns False if
        the packet mustn't be processed.

        Args:
            raw_data (Bytearray): received data.

        Returns:
            Boolean: ``True`` if the packet must be processed, ``False`` otherwise.
        """
        if raw_data[3] == ApiFrameType.RX_IO_64 and len(raw_data[14:-1]) < IOSample.min_io_sample_payload():
            return False
        return True

    def __try_add_remote_device(self, xbee_packet):
        """
        If the packet has information about a remote device, this method
        extracts that information from the packet, creates a remote device, and
        adds it (if not exist yet) to the network.

        Returns:
            :class:`.RemoteXBeeDevice`: the remote device extracted from the packet, `None`` if the packet has
                not information about a remote device.
        """
        remote = None
        x64, x16 = self.__get_remote_device_data_from_packet(xbee_packet)
        if x64 is not None or x16 is not None:
            remote = self.__xbee_device.get_network().add_if_not_exist(x64, x16)
        return remote

    @staticmethod
    def __is_special_explicit_packet(xbee_packet):
        """
        Checks if an explicit data packet is 'special'.

        'Special' means that this XBee has its API Output Mode distinct than Native (it's expecting
        explicit data packets), but some device has sent it a non-explicit data packet (TransmitRequest f.e.).
        In this case, this XBee will receive a explicit data packet with the following values:

            1. Source endpoint = 0xE8
            2. Destination endpoint = 0xE8
            3. Cluster ID = 0x0011
            4. Profile ID = 0xC105
        """
        if (xbee_packet.source_endpoint == 0xE8 and xbee_packet.dest_endpoint == 0xE8 and
                xbee_packet.cluster_id == 0x0011 and xbee_packet.profile_id == 0xC105):
            return True
        return False

    def __expl_to_no_expl(self, xbee_packet):
        """
        Creates a non-explicit data packet from the given explicit packet depending on
        this listener's XBee device protocol.

        Returns:
            :class:`.XBeeAPIPacket`: the proper receive packet depending on the current protocol and the
                available information (inside the packet).
        """
        x64addr = xbee_packet.x64bit_source_addr
        x16addr = xbee_packet.x16bit_source_addr
        if self.__xbee_device.get_protocol() == XBeeProtocol.RAW_802_15_4:
            if x64addr != XBee64BitAddress.UNKNOWN_ADDRESS:
                new_packet = RX64Packet(x64addr, 0, xbee_packet.receive_options, xbee_packet.rf_data)
            elif x16addr != XBee16BitAddress.UNKNOWN_ADDRESS:
                new_packet = RX16Packet(x16addr, 0, xbee_packet.receive_options, xbee_packet.rf_data)
            else:  # both address UNKNOWN
                new_packet = RX64Packet(x64addr, 0, xbee_packet.receive_options, xbee_packet.rf_data)
        else:
            new_packet = ReceivePacket(xbee_packet.x64bit_source_addr, xbee_packet.x16bit_source_addr,
                                       xbee_packet.receive_options, xbee_packet.rf_data)
        return new_packet

    def __add_packet_queue(self, xbee_packet):
        """
        Adds a packet to the queue. If the queue is full,
        the first packet of the queue is removed and the given
        packet is added.

        Args:
            xbee_packet (:class:`.XBeeAPIPacket`): the packet to be added.
        """
        # Data packets.
        if xbee_packet.get_frame_type() in [ApiFrameType.RECEIVE_PACKET, ApiFrameType.RX_64, ApiFrameType.RX_16]:
            if self.__data_xbee_queue.full():
                self.__data_xbee_queue.get()
            self.__data_xbee_queue.put_nowait(xbee_packet)
        # Explicit packets.
        if xbee_packet.get_frame_type() == ApiFrameType.EXPLICIT_RX_INDICATOR:
            if self.__explicit_xbee_queue.full():
                self.__explicit_xbee_queue.get()
            self.__explicit_xbee_queue.put_nowait(xbee_packet)
            # Check if the explicit packet is 'special'.
            if self.__is_special_explicit_packet(xbee_packet):
                # Create the non-explicit version of this packet and add it to the queue.
                self.__add_packet_queue(self.__expl_to_no_expl(xbee_packet))
        # IP packets.
        elif xbee_packet.get_frame_type() == ApiFrameType.RX_IPV4:
            if self.__ip_xbee_queue.full():
                self.__ip_xbee_queue.get()
            self.__ip_xbee_queue.put_nowait(xbee_packet)
        # Rest of packets.
        else:
            if self.__xbee_queue.full():
                self.__xbee_queue.get()
            self.__xbee_queue.put_nowait(xbee_packet)

    @staticmethod
    def __expl_to_message(remote, broadcast, xbee_packet):
        """
        Converts an explicit packet in an explicit message.

        Args:
            remote (:class:`.RemoteXBeeDevice`): the remote XBee device that sent the packet.
            broadcast (Boolean, optional, default=``False``): flag indicating whether the message is
                broadcast (``True``) or not (``False``). Optional.
            xbee_packet (:class:`.XBeeAPIPacket`): the packet to be converted.

        Returns:
            :class:`.ExplicitXBeeMessage`: the explicit message generated from the provided parameters.
        """
        return ExplicitXBeeMessage(xbee_packet.rf_data, remote, time.time(), xbee_packet.source_endpoint,
                                   xbee_packet.dest_endpoint, xbee_packet.cluster_id,
                                   xbee_packet.profile_id, broadcast)


class XBeeQueue(Queue):
    """
    This class represents an XBee queue.
    """

    def __init__(self, maxsize=10):
        """
        Class constructor. Instantiates a new :class:`.XBeeQueue` with the provided parameters.

        Args:
            maxsize (Integer, default: 10) the maximum size of the queue.
        """
        Queue.__init__(self, maxsize)

    def get(self, block=True, timeout=None):
        """
        Returns the first element of the queue if there is some
        element ready before timeout expires, in case of the timeout is not
        ``None``.

        If timeout is ``None``, this method is non-blocking. In this case, if there
        isn't any element available, it returns ``None``, otherwise it returns
        an :class:`.XBeeAPIPacket`.

        Args:
            block (Boolean): ``True`` to block during ``timeout`` waiting for a packet, ``False`` to not block.
            timeout (Integer, optional): timeout in seconds.

        Returns:
            :class:`.XBeeAPIPacket`: a packet if there is any packet available before ``timeout`` expires.
                If ``timeout`` is ``None``, the returned value may be ``None``.

        Raises:
            TimeoutException: if ``timeout`` is not ``None`` and there isn't any packet available
                before the timeout expires.
        """
        if timeout is None:
            try:
                xbee_packet = Queue.get(self, block=False)
            except (Empty, ValueError):
                xbee_packet = None
            return xbee_packet
        else:
            try:
                return Queue.get(self, True, timeout)
            except Empty:
                raise TimeoutException()

    def get_by_remote(self, remote_xbee_device, timeout=None):
        """
        Returns the first element of the queue that had been sent
        by ``remote_xbee_device``, if there is some in the specified timeout.

        If timeout is ``None``, this method is non-blocking. In this case, if there isn't
        any packet sent by ``remote_xbee_device`` in the queue, it returns ``None``,
        otherwise it returns an :class:`.XBeeAPIPacket`.

        Args:
            remote_xbee_device (:class:`.RemoteXBeeDevice`): the remote XBee device to get its firs element from queue.
            timeout (Integer, optional): timeout in seconds.

        Returns:
            :class:`.XBeeAPIPacket`: if there is any packet available before the timeout expires. If timeout is
                ``None``, the returned value may be ``None``.

        Raises:
            TimeoutException: if timeout is not ``None`` and there isn't any packet available that has
                been sent by ``remote_xbee_device`` before the timeout expires.
        """
        if timeout is None:
            with self.mutex:
                for xbee_packet in self.queue:
                    if self.__remote_device_match(xbee_packet, remote_xbee_device):
                        self.queue.remove(xbee_packet)
                        return xbee_packet
            return None
        else:
            xbee_packet = self.get_by_remote(remote_xbee_device, None)
            dead_line = time.time() + timeout
            while xbee_packet is None and dead_line > time.time():
                time.sleep(0.1)
                xbee_packet = self.get_by_remote(remote_xbee_device, None)
            if xbee_packet is None:
                raise TimeoutException()
            return xbee_packet

    def get_by_ip(self, ip_addr, timeout=None):
        """
        Returns the first IP data packet from the queue whose IP address
        matches the provided address.

        If timeout is ``None``, this method is non-blocking. In this case, if there isn't
        any packet sent by ``remote_xbee_device`` in the queue, it returns ``None``,
        otherwise it returns an :class:`.XBeeAPIPacket`.

        Args:
            ip_addr (:class:`ipaddress.IPv4Address`): The IP address to look for in the list of packets.
            timeout (Integer, optional): Timeout in seconds.

        Returns:
            :class:`.XBeeAPIPacket`: if there is any packet available before the timeout expires. If timeout is
                ``None``, the returned value may be ``None``.

        Raises:
            TimeoutException: if timeout is not ``None`` and there isn't any packet available that has
                been sent by ``remote_xbee_device`` before the timeout expires.
        """
        if timeout is None:
            with self.mutex:
                for xbee_packet in self.queue:
                    if self.__ip_addr_match(xbee_packet, ip_addr):
                        self.queue.remove(xbee_packet)
                        return xbee_packet
            return None
        else:
            xbee_packet = self.get_by_ip(ip_addr, None)
            dead_line = time.time() + timeout
            while xbee_packet is None and dead_line > time.time():
                time.sleep(0.1)
                xbee_packet = self.get_by_ip(ip_addr, None)
            if xbee_packet is None:
                raise TimeoutException()
            return xbee_packet

    def get_by_id(self, frame_id, timeout=None):
        """
        Returns the first packet from the queue whose frame ID
        matches the provided one.

        If timeout is ``None``, this method is non-blocking. In this case, if there isn't
        any received packet with the provided frame ID in the queue, it returns ``None``,
        otherwise it returns an :class:`.XBeeAPIPacket`.

        Args:
            frame_id (Integer): The frame ID to look for in the list of packets.
            timeout (Integer, optional): Timeout in seconds.

        Returns:
            :class:`.XBeeAPIPacket`: if there is any packet available before the timeout expires. If timeout is
                ``None``, the returned value may be ``None``.

        Raises:
            TimeoutException: if timeout is not ``None`` and there isn't any packet available that matches
            the provided frame ID before the timeout expires.
        """
        if timeout is None:
            with self.mutex:
                for xbee_packet in self.queue:
                    if xbee_packet.needs_id() and xbee_packet.frame_id == frame_id:
                        self.queue.remove(xbee_packet)
                        return xbee_packet
            return None
        else:
            xbee_packet = self.get_by_id(frame_id, None)
            dead_line = time.time() + timeout
            while xbee_packet is None and dead_line > time.time():
                time.sleep(0.1)
                xbee_packet = self.get_by_id(frame_id, None)
            if xbee_packet is None:
                raise TimeoutException()
            return xbee_packet

    def flush(self):
        """
        Clears the queue.
        """
        with self.mutex:
            self.queue.clear()

    @staticmethod
    def __remote_device_match(xbee_packet, remote_xbee_device):
        """
        Returns whether or not the source address of the provided XBee packet
        matches the address of the given remote XBee device.

        Args:
            xbee_packet (:class:`.XBeePacket`): The XBee packet to get the address to compare.
            remote_xbee_device (:class:`.RemoteXBeeDevice`): The remote XBee device to get the address to compare.

        Returns:
            Boolean: ``True`` if the remote device matches, ``False`` otherwise.
        """
        if xbee_packet.get_frame_type() == ApiFrameType.RECEIVE_PACKET:
            if xbee_packet.x64bit_source_addr == remote_xbee_device.get_64bit_addr():
                return True
            return xbee_packet.x16bit_source_addr == remote_xbee_device.get_16bit_addr()

        elif xbee_packet.get_frame_type() == ApiFrameType.REMOTE_AT_COMMAND_RESPONSE:
            if xbee_packet.x64bit_source_addr == remote_xbee_device.get_64bit_addr():
                return True
            return xbee_packet.x16bit_source_addr == remote_xbee_device.get_16bit_addr()

        elif xbee_packet.get_frame_type() == ApiFrameType.RX_16:
            return xbee_packet.x16bit_source_addr == remote_xbee_device.get_16bit_addr()

        elif xbee_packet.get_frame_type() == ApiFrameType.RX_64:
            return xbee_packet.x64bit_source_addr == remote_xbee_device.get_64bit_addr()

        elif xbee_packet.get_frame_type() == ApiFrameType.RX_IO_16:
            return xbee_packet.x16bit_source_addr == remote_xbee_device.get_16bit_addr()

        elif xbee_packet.get_frame_type() == ApiFrameType.RX_IO_64:
            return xbee_packet.x64bit_source_addr == remote_xbee_device.get_64bit_addr()

        elif xbee_packet.get_frame_type() == ApiFrameType.EXPLICIT_RX_INDICATOR:
            return xbee_packet.x64bit_source_addr == remote_xbee_device.get_64bit_addr()

        else:
            return False

    @staticmethod
    def __ip_addr_match(xbee_packet, ip_addr):
        """
        Returns whether or not the IP address of the XBee packet matches the
        provided one.

        Args:
            xbee_packet (:class:`.XBeePacket`): The XBee packet to get the address to compare.
            ip_addr (:class:`ipaddress.IPv4Address`): The IP address to be compared with the XBee packet's one.

        Returns:
            Boolean: ``True`` if the IP address matches, ``False`` otherwise.
        """
        return xbee_packet.get_frame_type() == ApiFrameType.RX_IPV4 and xbee_packet.source_address == ip_addr



# Copyright 2017, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from enum import Enum, unique

from digi.xbee.models.protocol import XBeeProtocol
from digi.xbee.util import utils


class ReceiveOptions(Enum):
    """
    This class lists all the possible options that have been set while
    receiving an XBee packet.

    The receive options are usually set as a bitfield meaning that the
    options can be combined using the '|' operand.
    """

    NONE = 0x00
    """
    No special receive options.
    """

    PACKET_ACKNOWLEDGED = 0x01
    """
    Packet was acknowledged.
    
    Not valid for WiFi protocol.
    """

    BROADCAST_PACKET = 0x02
    """
    Packet was a broadcast packet.
    
    Not valid for WiFi protocol.
    """

    APS_ENCRYPTED = 0x20
    """
    Packet encrypted with APS encryption.
    
    Only valid for ZigBee XBee protocol.
    """

    SENT_FROM_END_DEVICE = 0x40
    """
    Packet was sent from an end device (if known).
    
    Only valid for ZigBee XBee protocol.
    """


ReceiveOptions.__doc__ += utils.doc_enum(ReceiveOptions)


class TransmitOptions(Enum):
    """
    This class lists all the possible options that can be set while 
    transmitting an XBee packet.

    The transmit options are usually set as a bitfield meaning that the options 
    can be combined using the '|' operand.
    
    Not all options are available for all cases, that's why there are different 
    names with same values. In each moment, you must be sure that the option 
    your are going to use, is a valid option in your context.
    """

    NONE = 0x00
    """
    No special transmit options.
    """

    DISABLE_ACK = 0x01
    """
    Disables acknowledgments on all unicasts .

    Only valid for DigiMesh, 802.15.4 and Point-to-multipoint 
    protocols.
    """

    DISABLE_RETRIES_AND_REPAIR = 0x01
    """
    Disables the retries and router repair in the frame .

    Only valid for ZigBee protocol.
    """

    DONT_ATTEMPT_RD = 0x02
    """
    Doesn't attempt Route Discovery .

    Disables Route Discovery on all DigiMesh unicasts.

    Only valid for DigiMesh protocol.
    """

    USE_BROADCAST_PAN_ID = 0x04
    """
    Sends packet with broadcast {@code PAN ID}. Packet will be sent to all 
    devices in the same channel ignoring the {@code PAN ID}.

    It cannot be combined with other options.

    Only valid for 802.15.4 XBee protocol.
    """

    ENABLE_UNICAST_NACK = 0x04
    """
    Enables unicast NACK messages .

    NACK message is enabled on the packet.

    Only valid for DigiMesh 868/900 protocol.
    """

    ENABLE_UNICAST_TRACE_ROUTE = 0x04
    """
    Enables unicast trace route messages .

    Trace route is enabled on the packets.

    Only valid for DigiMesh 868/900 protocol.
    """

    ENABLE_MULTICAST = 0x08
    """
    Enables multicast transmission request.

    Only valid for ZigBee XBee protocol.
    """

    ENABLE_APS_ENCRYPTION = 0x20
    """
    Enables APS encryption, only if {@code EE=1} .

    Enabling APS encryption decreases the maximum number of RF payload 
    bytes by 4 (below the value reported by {@code NP}).

    Only valid for ZigBee XBee protocol.
    """

    USE_EXTENDED_TIMEOUT = 0x40
    """
    Uses the extended transmission timeout .

    Setting the extended timeout bit causes the stack to set the 
    extended transmission timeout for the destination address.

    Only valid for ZigBee XBee protocol.
    """

    POINT_MULTIPOINT_MODE = 0x40
    """
    Transmission is performed using point-to-Multipoint mode. 

    Only valid for DigiMesh 868/900 and Point-to-Multipoint 868/900 
    protocols.
    """

    REPEATER_MODE = 0x80
    """
    Transmission is performed using repeater mode .

    Only valid for DigiMesh 868/900 and Point-to-Multipoint 868/900 
    protocols.
    """

    DIGIMESH_MODE = 0xC0
    """
    Transmission is performed using DigiMesh mode .

    Only valid for DigiMesh 868/900 and Point-to-Multipoint 868/900 
    protocols.
    """


TransmitOptions.__doc__ += utils.doc_enum(TransmitOptions)


class RemoteATCmdOptions(Enum):
    """
    This class lists all the possible options that can be set while transmitting
    a remote AT Command.

    These options are usually set as a bitfield meaning that the options 
    can be combined using the '|' operand.
    """

    NONE = 0x00
    """
    No special transmit options
    """

    DISABLE_ACK = 0x01
    """
    Disables ACK
    """

    APPLY_CHANGES = 0x02
    """
    Applies changes in the remote device.

    If this option is not set, AC command must be sent before changes 
    will take effect.
    """

    EXTENDED_TIMEOUT = 0x40
    """
    Uses the extended transmission timeout

    Setting the extended timeout bit causes the stack to set the extended 
    transmission timeout for the destination address.

    Only valid for ZigBee XBee protocol.
    """


RemoteATCmdOptions.__doc__ += utils.doc_enum(RemoteATCmdOptions)


@unique
class SendDataRequestOptions(Enum):
    """
    Enumerates the different options for the :class:`.SendDataRequestPacket`.
    """
    OVERWRITE = (0, "Overwrite")
    ARCHIVE = (1, "Archive")
    APPEND = (2, "Append")
    TRANSIENT = (3, "Transient data (do not store)")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the SendDataRequestOptions element.

        Returns:
            Integer: the code of the SendDataRequestOptions element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the SendDataRequestOptions element.

        Returns:
            String: the description of the SendDataRequestOptions element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the send data request option for the given code.

        Args:
            code (Integer): the code of the send data request option to get.

        Returns:
            :class:`.FrameError`: the SendDataRequestOptions with the given code, ``None`` if there is not
                any send data request option with the provided code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return None

    code = property(__get_code)
    """Integer. The send data request option code."""

    description = property(__get_description)
    """String. The send data request option description."""


SendDataRequestOptions.lookupTable = {x.code: x for x in SendDataRequestOptions}
SendDataRequestOptions.__doc__ += utils.doc_enum(SendDataRequestOptions)


@unique
class DiscoveryOptions(Enum):
    """
    Enumerates the different options used in the discovery process.
    """

    APPEND_DD = (0x01, "Append device type identifier (DD)")
    """
    Append device type identifier (DD) to the discovery response.

    Valid for the following protocols:
      * DigiMesh
      * Point-to-multipoint (Digi Point)
      * ZigBee
    """

    DISCOVER_MYSELF = (0x02, "Local device sends response frame")
    """
    Local device sends response frame when discovery is issued.

    Valid for the following protocols:
      * DigiMesh
      * Point-to-multipoint (Digi Point)
      * ZigBee
      * 802.15.4
    """

    APPEND_RSSI = (0x04, "Append RSSI (of the last hop)")
    """
    Append RSSI of the last hop to the discovery response.

    Valid for the following protocols:
      * DigiMesh
      * Point-to-multipoint (Digi Point)
    """

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the ``DiscoveryOptions`` element.

        Returns:
            Integer: the code of the ``DiscoveryOptions`` element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the ``DiscoveryOptions`` element.

        Returns:
            String: the description of the ``DiscoveryOptions`` element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the ``DiscoveryOptions`` for the given code.

        Args:
            code (Integer): the code of the ``DiscoveryOptions`` to get.

        Returns:
            :class:`.FrameError`: the ``DiscoveryOptions`` with the given code, ``None`` if there is not
                any discovery option with the provided code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return None

    @staticmethod
    def calculate_discovery_value(protocol, options):
        """
        Calculates the total value of a combination of several options for the
        given protocol.

        Args:
            protocol (:class:`.XBeeProtocol`): the ``XBeeProtocol`` to calculate the value of all the given
                discovery options.
            options: collection of options to get the final value.

        Returns:
            Integer: The value to be configured in the module depending on the given
                collection of options and the protocol.
        """
        value = 0
        if protocol in [XBeeProtocol.ZIGBEE, XBeeProtocol.ZNET]:
            for op in options:
                if op == DiscoveryOptions.APPEND_RSSI:
                    continue
                value = value + op.code
        elif protocol in [XBeeProtocol.DIGI_MESH, XBeeProtocol.DIGI_POINT, XBeeProtocol.XLR, XBeeProtocol.XLR_DM]:
            for op in options:
                value = value + op.code
        else:
            if DiscoveryOptions.DISCOVER_MYSELF in options:
                value = 1  # This is different for 802.15.4.
        return value

    code = property(__get_code)
    """Integer. The discovery option code."""

    description = property(__get_description)
    """String. The discovery option description."""


DiscoveryOptions.lookupTable = {x.code: x for x in DiscoveryOptions}
DiscoveryOptions.__doc__ += utils.doc_enum(DiscoveryOptions)



# Copyright 2017, Digi International Inc.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from digi.xbee.devices import XBeeDevice
from digi.xbee.util import utils
from digi.xbee.models.address import XBee64BitAddress
from digi.xbee.models.status import PowerLevel

# TODO: Replace with the serial port where your local module is connected to. 
PORT = "COM1"
# TODO: Replace with the baud rate of your local module.
BAUD_RATE = 9600

PARAM_VALUE_PAN_ID = bytearray(b'\x01\x23')
PARAM_DESTINATION_ADDR = XBee64BitAddress.BROADCAST_ADDRESS
PARAM_POWER_LEVEL = PowerLevel.LEVEL_MEDIUM


def main():
    print(" +-----------------------------------------------------+")
    print(" | XBee Python Library Manage Common parameters Sample |")
    print(" +-----------------------------------------------------+\n")

    device = XBeeDevice(PORT, BAUD_RATE)

    try:
        device.open()

        print("Cached parameters\n" + "-" * 50)
        print("64-bit address:   %s" % device.get_64bit_addr())
        print("16-bit address:   %s" % device.get_16bit_addr())
        print("Node Identifier:  %s" % device.get_node_id())
        print("Firmware version: %s" % utils.hex_to_string(device.get_firmware_version()))
        print("Hardware version: %s" % device.get_hardware_version().description)

        print("")

        # Configure and read non-cached parameters.
        device.set_pan_id(PARAM_VALUE_PAN_ID)
        device.set_dest_address(PARAM_DESTINATION_ADDR)
        device.set_power_level(PARAM_POWER_LEVEL)

        print("Non-Cached parameters\n" + "-" * 50)
        print("PAN ID:           %s" % utils.hex_to_string(device.get_pan_id()))
        print("Dest address:     %s" % device.get_dest_address())
        print("Power level:      %s" % device.get_power_level().description)

    finally:
        if device is not None and device.is_open():
            device.close()


if __name__ == '__main__':
    main()



# Copyright 2017, Digi International Inc.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from digi.xbee.devices import WiFiDevice
from digi.xbee.models.mode import IPAddressingMode
from digi.xbee.models.status import ModemStatus

# TODO: Replace with the serial port where your local module is connected to.
PORT = "COM1"
# TODO: Replace with the baud rate of your local module.
BAUD_RATE = 9600
# TODO Fill with the SSID of the access point you want to connect to.
SSID = ""
# TODO Fill with the password of the access point you want to connect to.
PASSWORD = ""


def main():
    print(" +----------------------------------------------------+")
    print(" | XBee Python Library Connect to Access Point Sample |")
    print(" +----------------------------------------------------+\n")

    device = WiFiDevice(PORT, BAUD_RATE)

    try:
        device.open()

        if device.is_connected():
            device.disconnect()

        device.set_ip_addressing_mode(IPAddressingMode.DHCP)

        def modem_status_receive_callback(modem_status):
            print("Modem status: %s" % modem_status.description)
            if modem_status == ModemStatus.JOINED_NETWORK:
                print(">> Successfully connected to '%s'" % SSID)
            elif modem_status == ModemStatus.STATUS_DISASSOCIATED:
                print(">> Disconnected from the access point")

        device.add_modem_status_received_callback(modem_status_receive_callback)

        if not device.connect_by_ssid(SSID, PASSWORD):
            print(">> Error: could not connect to '%s'\n" % SSID)

        print("")
        print("  - IP addressing mode: %s" % device.get_ip_addressing_mode().description)
        print("  - IP address:         %s" % device.get_ip_addr().exploded)
        print("  - IP address mask:    %s" % device.get_mask_address().exploded)
        print("  - Gateway IP address: %s" % device.get_gateway_address().exploded)
        print("  - DNS address:        %s" % device.get_dns_address().exploded)
        print("")

    finally:
        if device is not None and device.is_open():
            device.close()


if __name__ == '__main__':
    main()



# Copyright 2017, Digi International Inc.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from digi.xbee.devices import XBeeDevice
from digi.xbee.util import utils

# TODO: Replace with the serial port where your local module is connected to. 
PORT = "COM1"
# TODO: Replace with the baud rate of your local module.
BAUD_RATE = 9600

PARAM_NODE_ID = "NI"
PARAM_PAN_ID = "ID"
PARAM_DEST_ADDRESS_H = "DH"
PARAM_DEST_ADDRESS_L = "DL"

PARAM_VALUE_NODE_ID = "Yoda"
PARAM_VALUE_PAN_ID = utils.hex_string_to_bytes("1234")
PARAM_VALUE_DEST_ADDRESS_H = utils.hex_string_to_bytes("00")
PARAM_VALUE_DEST_ADDRESS_L = utils.hex_string_to_bytes("FFFF")


def main():
    print(" +-----------------------------------------------+")
    print(" | XBee Python Library Set/Get parameters Sample |")
    print(" +-----------------------------------------------+\n")

    device = XBeeDevice(PORT, BAUD_RATE)

    try:
        device.open()

        # Set parameters.
        device.set_parameter(PARAM_NODE_ID, bytearray(PARAM_VALUE_NODE_ID, 'utf8'))
        device.set_parameter(PARAM_PAN_ID, PARAM_VALUE_PAN_ID)
        device.set_parameter(PARAM_DEST_ADDRESS_H, PARAM_VALUE_DEST_ADDRESS_H)
        device.set_parameter(PARAM_DEST_ADDRESS_L, PARAM_VALUE_DEST_ADDRESS_L)

        # Get parameters.
        print("Node ID:                     %s" % device.get_parameter(PARAM_NODE_ID).decode())
        print("PAN ID:                      %s" % utils.hex_to_string(device.get_parameter(PARAM_PAN_ID)))
        print("Destination address high:    %s" % utils.hex_to_string(device.get_parameter(PARAM_DEST_ADDRESS_H)))
        print("Destination address low:     %s" % utils.hex_to_string(device.get_parameter(PARAM_DEST_ADDRESS_L)))

        print("")
        print("All parameters were set correctly!")

    finally:
        if device is not None and device.is_open():
            device.close()


if __name__ == '__main__':
    main()



# Copyright 2017, Digi International Inc.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from digi.xbee.devices import XBeeDevice
from digi.xbee.io import IOLine, IOMode
import time
import threading

# TODO: Replace with the serial port where your local module is connected to. 
PORT = "COM1"
# TODO: Replace with the baud rate of your local module.
BAUD_RATE = 9600

IO_LINE_IN = IOLine.DIO3_AD3
IO_LINE_OUT = IOLine.DIO4_AD4


def main():
    print(" +----------------------------------------------+")
    print(" | XBee Python Library Get/Set Local DIO Sample |")
    print(" +----------------------------------------------+\n")

    stop = False
    th = None

    device = XBeeDevice(PORT, BAUD_RATE)

    try:
        device.open()

        def io_detection_callback():
            while not stop:
                # Read the digital value from the input line.
                io_value = device.get_dio_value(IO_LINE_IN)
                print("%s: %s" % (IO_LINE_IN, io_value))

                # Set the previous value to the output line.
                device.set_dio_value(IO_LINE_OUT, io_value)

                time.sleep(1)

        th = threading.Thread(target=io_detection_callback)

        device.set_io_configuration(IO_LINE_IN, IOMode.DIGITAL_IN)
        device.set_io_configuration(IO_LINE_OUT, IOMode.DIGITAL_OUT_LOW)

        time.sleep(1)
        th.start()

        input()

    finally:
        stop = True
        if th is not None and th.is_alive():
            th.join()
        if device is not None and device.is_open():
            device.close()


if __name__ == '__main__':
    main()



# Copyright 2017, 2018, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from digi.xbee.packets.base import *
from digi.xbee.packets.cellular import *
from digi.xbee.packets.common import *
from digi.xbee.packets.devicecloud import *
from digi.xbee.packets.network import *
from digi.xbee.packets.raw import *
from digi.xbee.packets.wifi import *
from digi.xbee.packets.aft import ApiFrameType
from digi.xbee.models.mode import OperatingMode


"""
This module provides functionality to build XBee packets from
bytearray returning the appropriate XBeePacket subclass.

All the API and API2 logic is already included so all packet reads are
independent of the XBee operating mode.

Two API modes are supported and both can be enabled using the ``AP``
(API Enable) command::

    API1 - API Without Escapes
    The data frame structure is defined as follows:

      Start Delimiter          Length                   Frame Data                   Checksum
          (Byte 1)            (Bytes 2-3)               (Bytes 4-n)                (Byte n + 1)
    +----------------+  +-------------------+  +--------------------------- +  +----------------+
    |      0x7E      |  |   MSB   |   LSB   |  |   API-specific Structure   |  |     1 Byte     |
    +----------------+  +-------------------+  +----------------------------+  +----------------+
                   MSB = Most Significant Byte, LSB = Least Significant Byte


API2 - API With Escapes
The data frame structure is defined as follows::

      Start Delimiter          Length                   Frame Data                   Checksum
          (Byte 1)            (Bytes 2-3)               (Bytes 4-n)                (Byte n + 1)
    +----------------+  +-------------------+  +--------------------------- +  +----------------+
    |      0x7E      |  |   MSB   |   LSB   |  |   API-specific Structure   |  |     1 Byte     |
    +----------------+  +-------------------+  +----------------------------+  +----------------+
                        \___________________________________  _________________________________/
                                                            \/
                                                Characters Escaped If Needed

                   MSB = Most Significant Byte, LSB = Least Significant Byte


When sending or receiving an API2 frame, specific data values must be
escaped (flagged) so they do not interfere with the data frame sequencing.
To escape an interfering data byte, the byte 0x7D is inserted before
the byte to be escaped XOR'd with 0x20.

The data bytes that need to be escaped:

- ``0x7E`` - Frame Delimiter - :attr:`.SpecialByte.
- ``0x7D`` - Escape
- ``0x11`` - XON
- ``0x13`` - XOFF

The length field has a two-byte value that specifies the number of
bytes that will be contained in the frame data field. It does not include the
checksum field.

The frame data  forms an API-specific structure as follows::

      Start Delimiter          Length                   Frame Data                   Checksum
          (Byte 1)            (Bytes 2-3)               (Bytes 4-n)                (Byte n + 1)
    +----------------+  +-------------------+  +--------------------------- +  +----------------+
    |      0x7E      |  |   MSB   |   LSB   |  |   API-specific Structure   |  |     1 Byte     |
    +----------------+  +-------------------+  +----------------------------+  +----------------+
                                               /                                                 \\
                                              /  API Identifier        Identifier specific data   \\
                                              +------------------+  +------------------------------+
                                              |       cmdID      |  |           cmdData            |
                                              +------------------+  +------------------------------+


The cmdID frame (API-identifier) indicates which API messages
will be contained in the cmdData frame (Identifier-specific data).

To unit_test data integrity, a checksum is calculated and verified on
non-escaped data.

.. seealso::
   | :class:`.XBeePacket`
   | :class:`.OperatingMode`
"""


def build_frame(packet_bytearray, operating_mode=OperatingMode.API_MODE):
    """
    Creates a packet from raw data.
    
    Args:
        packet_bytearray (Bytearray): the raw data of the packet to build.
        operating_mode (:class:`.OperatingMode`): the operating mode in which the raw data has been captured.

    .. seealso::
       | :class:`.OperatingMode`
    """
    frame_type = ApiFrameType.get(packet_bytearray[3])

    if frame_type == ApiFrameType.GENERIC:
        return GenericXBeePacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.AT_COMMAND:
        return ATCommPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.AT_COMMAND_QUEUE:
        return ATCommQueuePacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.AT_COMMAND_RESPONSE:
        return ATCommResponsePacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.RECEIVE_PACKET:
        return ReceivePacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.RX_64:
        return RX64Packet.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.RX_16:
        return RX16Packet.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.REMOTE_AT_COMMAND_REQUEST:
        return RemoteATCommandPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.REMOTE_AT_COMMAND_RESPONSE:
        return RemoteATCommandResponsePacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.TRANSMIT_REQUEST:
        return TransmitPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.TRANSMIT_STATUS:
        return TransmitStatusPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.MODEM_STATUS:
        return ModemStatusPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.TX_STATUS:
        return TXStatusPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.RX_IO_16:
        return RX16IOPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.RX_IO_64:
        return RX64IOPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.IO_DATA_SAMPLE_RX_INDICATOR:
        return IODataSampleRxIndicatorPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.EXPLICIT_ADDRESSING:
        return ExplicitAddressingPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.EXPLICIT_RX_INDICATOR:
        return ExplicitRXIndicatorPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.TX_SMS:
        return TXSMSPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.TX_IPV4:
        return TXIPv4Packet.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.RX_SMS:
        return RXSMSPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.RX_IPV4:
        return RXIPv4Packet.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.REMOTE_AT_COMMAND_REQUEST_WIFI:
        return RemoteATCommandWifiPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.SEND_DATA_REQUEST:
        return SendDataRequestPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.DEVICE_RESPONSE:
        return DeviceResponsePacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.REMOTE_AT_COMMAND_RESPONSE_WIFI:
        return RemoteATCommandResponseWifiPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.IO_DATA_SAMPLE_RX_INDICATOR_WIFI:
        return IODataSampleRxIndicatorWifiPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.SEND_DATA_RESPONSE:
        return SendDataResponsePacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.DEVICE_REQUEST:
        return DeviceRequestPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.DEVICE_RESPONSE_STATUS:
        return DeviceResponseStatusPacket.create_packet(packet_bytearray, operating_mode)

    elif frame_type == ApiFrameType.FRAME_ERROR:
        return FrameErrorPacket.create_packet(packet_bytearray, operating_mode)

    else:
        return UnknownXBeePacket.create_packet(packet_bytearray, operating_mode)



# Copyright 2017, Digi International Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from enum import Enum, unique
from digi.xbee.util import utils


@unique
class HardwareVersion(Enum):
    """
    This class lists all hardware versions.

    | Inherited properties:
    |     **name** (String): The name of this HardwareVersion.
    |     **value** (Integer): The ID of this HardwareVersion.
    """
    X09_009 = (0x01, "X09-009")
    X09_019 = (0x02, "X09-019")
    XH9_009 = (0x03, "XH9-009")
    XH9_019 = (0x04, "XH9-019")
    X24_009 = (0x05, "X24-009")
    X24_019 = (0x06, "X24-019")
    X09_001 = (0x07, "X09-001")
    XH9_001 = (0x08, "XH9-001")
    X08_004 = (0x09, "X08-004")
    XC09_009 = (0x0A, "XC09-009")
    XC09_038 = (0x0B, "XC09-038")
    X24_038 = (0x0C, "X24-038")
    X09_009_TX = (0x0D, "X09-009-TX")
    X09_019_TX = (0x0E, "X09-019-TX")
    XH9_009_TX = (0x0F, "XH9-009-TX")
    XH9_019_TX = (0x10, "XH9-019-TX")
    X09_001_TX = (0x11, "X09-001-TX")
    XH9_001_TX = (0x12, "XH9-001-TX")
    XT09B_XXX = (0x13, "XT09B-xxx (Attenuator version)")
    XT09_XXX = (0x14, "XT09-xxx")
    XC08_009 = (0x15, "XC08-009")
    XC08_038 = (0x16, "XC08-038")
    XB24_AXX_XX = (0x17, "XB24-Axx-xx")
    XBP24_AXX_XX = (0x18, "XBP24-Axx-xx")
    XB24_BXIX_XXX = (0x19, "XB24-BxIx-xxx and XB24-Z7xx-xxx")
    XBP24_BXIX_XXX = (0x1A, "XBP24-BxIx-xxx and XBP24-Z7xx-xxx")
    XBP09_DXIX_XXX = (0x1B, "XBP09-DxIx-xxx Digi Mesh")
    XBP09_XCXX_XXX = (0x1C, "XBP09-XCxx-xxx: S3 XSC Compatibility")
    XBP08_DXXX_XXX = (0x1D, "XBP08-Dxx-xxx 868MHz")
    XBP24B = (0x1E, "XBP24B: Low cost ZB PRO and PLUS S2B")
    XB24_WF = (0x1F, "XB24-WF: XBee 802.11 (Redpine module)")
    AMBER_MBUS = (0x20, "??????: M-Bus module made by Amber")
    XBP24C = (0x21, "XBP24C: XBee PRO SMT Ember 357 S2C PRO")
    XB24C = (0x22, "XB24C: XBee SMT Ember 357 S2C")
    XSC_GEN3 = (0x23, "XSC_GEN3: XBP9 XSC 24 dBm")
    SRD_868_GEN3 = (0x24, "SDR_868_GEN3: XB8 12 dBm")
    ABANDONATED = (0x25, "Abandonated")
    SMT_900LP = (0x26, "900LP (SMT): 900LP on 'S8 HW'")
    WIFI_ATHEROS = (0x27, "WiFi Atheros (TH-DIP) XB2S-WF")
    SMT_WIFI_ATHEROS = (0x28, "WiFi Atheros (SMT) XB2B-WF")
    SMT_475LP = (0x29, "475LP (SMT): Beta 475MHz")
    XBEE_CELL_TH = (0x2A, "XBee-Cell (TH): XBee Cellular")
    XLR_MODULE = (0x2B, "XLR Module")
    XB900HP_NZ = (0x2C, "XB900HP (New Zealand): XB9 NZ HW/SW")
    XBP24C_TH_DIP = (0x2D, "XBP24C (TH-DIP): XBee PRO DIP")
    XB24C_TH_DIP = (0x2E, "XB24C (TH-DIP): XBee DIP")
    XLR_BASEBOARD = (0x2F, "XLR Baseboard")
    XBP24C_S2C_SMT = (0x30, "XBee PRO SMT")
    SX_PRO = (0x31, "SX Pro")
    S2D_SMT_PRO = (0x32, "XBP24D: S2D SMT PRO")
    S2D_SMT_REG = (0x33, "XB24D: S2D SMT Reg")
    S2D_TH_PRO = (0x34, "XBP24D: S2D TH PRO")
    S2D_TH_REG = (0x35, "XB24D: S2D TH Reg")
    SX = (0x3E, "SX")
    XTR = (0x3F, "XTR")
    CELLULAR_CAT1_LTE_VERIZON = (0x40, "XBee Cellular Cat 1 LTE Verizon")
    XBEE3 = (0x41, "XBEE3")
    XBEE3_SMT = (0x42, "XBEE3 SMT")
    XBEE3_TH = (0x43, "XBEE3 TH")
    CELLULAR_3G = (0x44, "XBee Cellular 3G")
    XB8X = (0x45, "XB8X")
    CELLULAR_LTE_VERIZON = (0x46, "XBee Cellular LTE-M Verizon")
    CELLULAR_LTE_ATT = (0x47, "XBee Cellular LTE-M AT&T")
    CELLULAR_NBIOT_EUROPE = (0x48, "XBee Cellular NBIoT Europe")
    CELLULAR_3_CAT1_LTE_ATT = (0x49, "XBee Cellular 3 Cat 1 LTE AT&T")
    CELLULAR_3_LTE_M_VERIZON = (0x4A, "XBee Cellular 3 LTE-M Verizon")
    CELLULAR_3_LTE_M_ATT = (0x4B, "XBee Cellular 3 LTE-M AT&T")

    def __init__(self, code, description):
        self.__code = code
        self.__description = description

    def __get_code(self):
        """
        Returns the code of the HardwareVersion element.

        Returns:
            Integer: the code of the HardwareVersion element.
        """
        return self.__code

    def __get_description(self):
        """
        Returns the description of the HardwareVersion element.

        Returns:
            String: the description of the HardwareVersion element.
        """
        return self.__description

    @classmethod
    def get(cls, code):
        """
        Returns the HardwareVersion for the given code.

        Args:
            code (Integer): the code of the hardware version to get.

        Returns:
            :class:`HardwareVersion`: the HardwareVersion with the given code, ``None`` if there is not a
                HardwareVersion with that code.
        """
        try:
            return cls.lookupTable[code]
        except KeyError:
            return None

    code = property(__get_code)
    """Integer. The hardware version code."""

    description = property(__get_description)
    """String. The hardware version description."""


# Class variable of HardwareVersion. Dictionary <Integer, HardwareVersion> for
# search HardwareVersion by code.
HardwareVersion.lookupTable = {x.code: x for x in HardwareVersion}
HardwareVersion.__doc__ += utils.doc_enum(HardwareVersion)
